import logging
import os
import datetime
import json
import re
import requests
import base64

from typing import Dict, Union, Tuple
from pathlib import PurePath
from enum import Enum

from aceutils.logger import LoggingAdapter
from aceutils.file_util import load_json
from acedeploy.core.model_instance_objects import InstanceObject, ColumnInstance
from acedeploy.core.model_sql_entities import DbActionType, DbObjectType
import aceutils.string_util as string_util

logger = logging.getLogger(__name__)
log = LoggingAdapter(logger)

class PolicyAssignmentObjectTypes(Enum):
    """
    Enum for valid object types for policy assignments. 
    Subset of DbObjectType.
    Policy handling type PolicyHandlingTypes.NEW_OBJECT is supported for these objects.
    """
    VIEW = "VIEW"
    TABLE = "TABLE"
    DYNAMICTABLE = "DYNAMICTABLE"
    TAG = "TAG"

class PolicyAssignmentLevel(Enum):
    """
    Enum for valid policy assignment levels.
    """
    TABLE_LEVEL_ASSIGNMENT = "TABLE_LEVEL_ASSIGNMENT"
    SCHEMA_LEVEL_ASSIGNMENT = "SCHEMA_LEVEL_ASSIGNMENT"

class PolicyAssignmentType(Enum):
    """
    Enum for valid policy assignment types.
    """
    TABLECOLUMNS = "TABLE_COLUMNS"
    VIEWCOLUMNS = "VIEW_COLUMNS"
    DYNAMICTABLECOLUMNS = "DYNAMICTABLE_COLUMNS"
    VIEWS = "VIEWS"
    TABLES = "TABLES"
    DYNAMICTABLES = "DYNAMICTABLES"
    TAGS = "TAGS"

    @staticmethod
    def get_object_type_from_assignment_type(assignment_type) -> PolicyAssignmentObjectTypes:
        """
        Returns the corresponding object type for a policy assignment type.
        """
        if assignment_type.upper() == PolicyAssignmentType.TABLECOLUMNS.value:
            return PolicyAssignmentObjectTypes.TABLE
        if assignment_type.upper() == PolicyAssignmentType.VIEWCOLUMNS.value:
            return PolicyAssignmentObjectTypes.VIEW
        if assignment_type.upper() == PolicyAssignmentType.DYNAMICTABLECOLUMNS.value:
            return PolicyAssignmentObjectTypes.DYNAMICTABLE
        if assignment_type.upper() == PolicyAssignmentType.VIEWS.value:
            return PolicyAssignmentObjectTypes.VIEW
        if assignment_type.upper() == PolicyAssignmentType.TABLES.value:
            return PolicyAssignmentObjectTypes.TABLE
        if assignment_type.upper() == PolicyAssignmentType.TAGS.value:
            return PolicyAssignmentObjectTypes.TAG
        if assignment_type.upper() == PolicyAssignmentType.DYNAMICTABLES.value:
            return PolicyAssignmentObjectTypes.DYNAMICTABLE
        else:
            raise ValueError(f"Policy assignment type {assignment_type} not supported")
        
    @staticmethod
    def get_policy_assignment_level(policy_assignment_type) -> PolicyAssignmentLevel:
        """
        Returns object level of the desired policy assignment type.
        Currently supported are table-level objects (like columns) and schema-level objects like tables.
        """
        policy_assignment_level = {
            PolicyAssignmentType.TABLECOLUMNS: PolicyAssignmentLevel.TABLE_LEVEL_ASSIGNMENT,
            PolicyAssignmentType.VIEWCOLUMNS: PolicyAssignmentLevel.TABLE_LEVEL_ASSIGNMENT,
            PolicyAssignmentType.DYNAMICTABLECOLUMNS: PolicyAssignmentLevel.TABLE_LEVEL_ASSIGNMENT,
            PolicyAssignmentType.VIEWS: PolicyAssignmentLevel.SCHEMA_LEVEL_ASSIGNMENT,
            PolicyAssignmentType.TABLES: PolicyAssignmentLevel.SCHEMA_LEVEL_ASSIGNMENT,
            PolicyAssignmentType.TAGS: PolicyAssignmentLevel.SCHEMA_LEVEL_ASSIGNMENT,
            PolicyAssignmentType.DYNAMICTABLES: PolicyAssignmentLevel.SCHEMA_LEVEL_ASSIGNMENT,
        }

        return policy_assignment_level[policy_assignment_type]
    
class PolicyHandlingTypes(Enum):
    """
    Enum for valid policy handling types for which policy assignments are managed during deployment.
    """
    NEW_OBJECT = "NEW_OBJECT" # only when object type in PolicyAssignmentObjectTypes
    ALTER_VIEW = "ALTER_VIEW"
    ALTER_DYNAMICTABLE = "ALTER_DYNAMICTABLE" # only when alter query_text or alter refresh_mode
    ADD_TABLECOLUMN = "ADD_TABLECOLUMN"



class PolicyAssignmentAlterObjectTypes(Enum):
    """
    Enum for object types for which policy assignments are lost on deployment when altered. 
    Subset of PolicyAssignmentObjectTypes.
    Policy handling type can be retrieved with get_policy_handling_type.
    """
    VIEW = "VIEW"
    TABLE = "TABLE" # only when columns_to_add
    DYNAMICTABLE = "DYNAMICTABLE" # only when alter query_text or alter refresh_mode
    
    @staticmethod
    def get_policy_handling_type(policy_assignment_alter_object_type) -> PolicyHandlingTypes:
        """
        Returns the policy handling type for a given member of PolicyAssignmentAlterObjectTypes.
        """
        policy_handling_types = {
            PolicyAssignmentAlterObjectTypes.VIEW: PolicyHandlingTypes.ALTER_VIEW,
            PolicyAssignmentAlterObjectTypes.TABLE: PolicyHandlingTypes.ADD_TABLECOLUMN,
            PolicyAssignmentAlterObjectTypes.DYNAMICTABLE: PolicyHandlingTypes.ALTER_DYNAMICTABLE
        }
        return policy_handling_types[policy_assignment_alter_object_type]

def get_policy_assignments_info_from_object(
        object_schema: str, 
        object_name: str, 
        object_type: DbObjectType, 
        action_type: DbActionType, 
        policy_assignments_info: dict) -> list[Dict]:
    """
    Given an object action for which the policy assignments need to be handled, 
    get the policy assignments info for a specific object from the policy assignments info for all objects.
    """
    object_identifier = f"{object_schema}.{object_name}"
    policy_assignments_info_from_object={}
    
    if (
        (
            ( object_type.name in PolicyAssignmentAlterObjectTypes.__members__ and action_type.name == 'ALTER')
            or (object_type.name in PolicyAssignmentObjectTypes.__members__ and action_type.name == 'ADD')
        )
        and object_type.name in policy_assignments_info 
        and object_identifier in policy_assignments_info[object_type.name]
        ):
        if "desired_instance" not in policy_assignments_info[object_type.name][object_identifier]:
            raise ValueError(f"No desired_instance found in policy_assignments_info for object {object_identifier} with object type {object_type.name}")
        elif "assignments" in policy_assignments_info[object_type.name][object_identifier]["desired_instance"]:
            policy_assignments_info_from_object['assignments']=policy_assignments_info[object_type.name][object_identifier]["desired_instance"]["assignments"]
            policy_assignments_info_from_object['policy_handling_type']=policy_assignments_info[object_type.name][object_identifier]['policy_handling_type']
            policy_assignments_info_from_object['object_identifier']=object_identifier
    return policy_assignments_info_from_object

def create_columns_string_of_dynamictable_with_policy_assignments(columns:list[ColumnInstance], policy_assignments_of_object:list[Dict], always_create: bool = False) -> Tuple[str, str]:
    """
    Create a string defining the columns of a dynamic table containing CMP assignments on those columns.
    """
    if policy_assignments_of_object:
        policy_assignments_on_columns=[policy_assignment for policy_assignment in policy_assignments_of_object if policy_assignment["assignment_type"] == 'dynamictable_columns']
    else:
        policy_assignments_on_columns=[]
    
    #sort columns by ordinal position
    columns=sorted(columns, key=lambda x: x.ordinal_position)

    if policy_assignments_on_columns or always_create:
        columns_string='('

        for n_column, column in enumerate(columns):
            if n_column+1 != column.ordinal_position:
                raise ValueError(f"While creating a columns-string for the dynamic table {column.object_schema}.{column.object_name} the ordinal position of the column {column.column_name} does not match the position in the columns string.")
            column_identifier= f"{column.object_schema}.{column.object_name}.{column.column_name}"

            policy_assignment_on_column=next((assignment for assignment in policy_assignments_on_columns if assignment["assignment"] == column_identifier), None)
            if policy_assignment_on_column:
                if policy_assignment_on_column["argument_columns"]:
                    using_string=f"{column.column_name}"
                    for arg_column in policy_assignment_on_column["argument_columns"]:
                        using_string = f"{using_string}, {arg_column}"
                    using_string = f"USING ({using_string})"
                else:
                    using_string=''
                column_definition= f'{column.column_name} WITH MASKING POLICY {policy_assignment_on_column["policy_database"]}.{policy_assignment_on_column["policy_schema"]}.{policy_assignment_on_column["policy"]} {using_string}, '
            else:
                column_definition=f"{column.column_name}, "
            columns_string=columns_string + column_definition

        columns_string=columns_string[:-2] + ')'
    else:
        columns_string = '' 

    return columns_string

def create_columns_string_of_view_with_policy_assignments(columns:list[ColumnInstance], policy_assignments_of_object:list[Dict], always_create: bool = False) -> Tuple[str, str]:
    """
    Create a string defining the columns of a view containing CMP assignments on those columns.
    """
    if policy_assignments_of_object:
        policy_assignments_on_columns=[policy_assignment for policy_assignment in policy_assignments_of_object if policy_assignment["assignment_type"] == 'view_columns']
    else:
        policy_assignments_on_columns=[]
    
    #sort columns by ordinal position
    columns=sorted(columns, key=lambda x: x.ordinal_position)

    if policy_assignments_on_columns or always_create:
        columns_string='('
        columns_string_without_database_reference='('
        for n_column, column in enumerate(columns):
            if n_column+1 != column.ordinal_position:
                raise ValueError(f"While creating a columns-string for the view {column.object_schema}.{column.object_name} the ordinal position of the column {column.column_name} does not match the position in the columns string.")
            column_identifier= f"{column.object_schema}.{column.object_name}.{column.column_name}"
            if column.comment:
                comment_string= f"COMMENT '{column.comment}'"
            else:
                comment_string=''

            policy_assignment_on_column=next((assignment for assignment in policy_assignments_on_columns if assignment["assignment"] == column_identifier), None)
            if policy_assignment_on_column:
                if policy_assignment_on_column["argument_columns"]:
                    using_string=f"{column.column_name}"
                    for arg_column in policy_assignment_on_column["argument_columns"]:
                        using_string = f"{using_string}, {arg_column}"
                    using_string = f"USING ({using_string})"
                else:
                    using_string=''
                column_definition= f'{column.column_name} WITH MASKING POLICY {policy_assignment_on_column["policy_database"]}.{policy_assignment_on_column["policy_schema"]}.{policy_assignment_on_column["policy"]} {using_string} {comment_string}, '
                column_definition_without_database_reference= f'{column.column_name} WITH MASKING POLICY {policy_assignment_on_column["policy_schema"]}.{policy_assignment_on_column["policy"]} {using_string} {comment_string}, '
            else:
                column_definition=f"{column.column_name} {comment_string}, "
                column_definition_without_database_reference=f"{column.column_name} {comment_string}, "
            columns_string=columns_string + column_definition
            columns_string_without_database_reference=columns_string_without_database_reference + column_definition_without_database_reference

        columns_string=columns_string[:-2] + ')'
        columns_string_without_database_reference=columns_string_without_database_reference[:-2] + ')'
    else:
        columns_string = '' 
        columns_string_without_database_reference = ''

    return columns_string, columns_string_without_database_reference

def check_object_statement_for_columns_string(object_statement: str, columns: list[ColumnInstance]) -> Union[str, None]:
    """
    Check if the DDL statement of an view contains a columns_string. 
    Returns the columns string or None if no columns string was found.
    """
    columns_string=None

    for column in columns:
        if column.ordinal_position != 1:
            continue
        else:
            column_name = column.column_name
            columns_string=extract_columns_string(object_statement, column_name)
            
    return columns_string

def extract_columns_string(object_statement: str, column: str) -> Union[str, None]:
    """
    Extracts the substring between the first open bracket and the corresponding closing bracket
    that contains the given column name in the object statement.
    Ignores matches in comments.

    Args:
        s (str): The input string to extract the substring from.
        column (str): The name of the column to search for.

    Returns:
        Union[str, None]: The extracted substring, or None if the pattern is not found.
    """
    if not column:
        raise ValueError(f"While trying to extract columns string from view definition missing information about column with ordinal position 1. View definition: {object_statement}")
     
    regex = r'\(\s*' + re.escape(column) + r'\s*'
    stack = []
    ignore = False
    for i, char in enumerate(object_statement):
        if char == "'" and (i == 0 or object_statement[i-1] != "\\"):
            ignore = not ignore
        if ignore:
            continue
        if char == '(':
            stack.append('(')
        elif char == ')':
            stack.pop()
        if not stack and re.search(regex, object_statement[:i], re.IGNORECASE):
            start = re.search(regex, object_statement[:i], re.IGNORECASE).start()
            end = i
            return object_statement[start:end+1]
    return None

def replace_ignore(input_string: str, replace_pattern: str, replace_with: str, ignore_substring: str=None) -> str:
    """
    Replaces all occurences of a pattern in a string but ignores all occurences in a specified sub-string.
    This function is case insensitive.

    Args:
        input_string (str): The input string .
        replace_pattern (str): The pattern which should be replaced.
        replace_with (str): The string which should be inserted during replacement.
        ignore_substring (str): The sub-string which should be ignored.
    """
    if ignore_substring:
        replace_pattern=replace_pattern+fr'|{re.escape(ignore_substring)}'
    # TODO move function to string utils
    output_string = re.sub(
        replace_pattern,
        lambda match: match.group() if match.group() == ignore_substring else replace_with,
        input_string,
        flags=re.I
    )

    return output_string

def remove_comment_from_view_statement(object_statement: str, view_comment: str , columns_string: str=None) -> Union[str, None]:
    """
    Remove the comment string (e.g. "COMMENT= 'This is a View'") of an object statement and replace it with a whitespace.
    Ignores comments in the columns definition.
    Ignores SQL-comments.

    Args:
        object_statement (str): The input string representing the object_statement.
        comment_string (str): The comment string of the object statement.
        columns_string (str): The columns definition of the object_statement.

    Returns:
        Union[str, None]: The object_statement without the columns string.
    """
    #TODO does this function need to ignore occurences of the comment_string in SQL comments (--, //, /* */) ?
    if view_comment:
        pattern=fr"\s*comment\s*=\s*'{re.escape(view_comment)}'\s*"
        object_statement_without_comment_string = replace_ignore(input_string=object_statement, replace_pattern=pattern, replace_with=' ', ignore_substring=columns_string)
    else:
        object_statement_without_comment_string = object_statement
    return object_statement_without_comment_string

def remove_database_name_from_view_statement(view_definition: str) -> str:
    """
    Remove the database reference from the view identifier in a Snowflake View DDL statement.
    """
    pattern = fr"\s*CREATE\s+(?:OR\s+REPLACE\s+)?(?:SECURE\s+)?(?:RECURSIVE\s+)?VIEW\s+(?:IF\s+NOT\s+EXISTS\s+)?{string_util.REGEX_OBJ_IDENTIFIER_3}"

    m = re.match(pattern, view_definition, re.I)

    if m:
        view_definition=view_definition.replace(m.group('database'), "", 1)

    return view_definition

def prepare_view_statement_comparison_independent_of_columns_string(view_definition: str, columns: list[ColumnInstance], comment: str, policy_assignments_info_from_object: dict = {}) -> Tuple[str, str, str]:
    """
    Prepares the comparison of the current view definition and the desired view definition during for ALTER VIEW ACTIONS.
    If the current view definition contains a columns string, creates an alternative version without the columns string.
    If the current view definition does not contain a columns string, creates an alternative version with a columns string.

    Notes:
        - Removes the database reference from the view definition.
        - Adds OR REPLACE if not already in the view definiton.
        - Removes SQL-comments at the beginning of the view definition.

    Background: 
    When querying the definition of a view (that was created without a columns definition) the columns definition is sometimes omitted 
    (e.g. INFORMATION_SCHEMA.VIEWS, or SHOW VIEWS) and sometimes included (e.g. SELECT GET_DDL) depending on the query.
    """

    view_definition = string_util.remove_comment(view_definition)

    #TODO should this function be moved to model_object_action_entitites?
    current_instance_view_definition_with_columns_string = 'current view definition with columns string not defined'
    current_instance_view_definition_without_columns_string = 'current view definition without columns string not defined'

    current_instance_view_definition = remove_database_name_from_view_statement(string_util.add_create_or_replace(view_definition) )
    current_instance_columns_string = check_object_statement_for_columns_string(object_statement=view_definition, columns=columns)

    current_instance_view_definition = remove_comment_from_view_statement(current_instance_view_definition, comment, current_instance_columns_string)
    
    if not current_instance_columns_string:
        current_instance_columns_string, _ = create_columns_string_of_view_with_policy_assignments(columns=columns, policy_assignments_of_object={}, always_create = True)
        if current_instance_columns_string:
            current_instance_view_definition_with_columns_string = string_util.add_create_or_replace(add_string_to_object_ddl(object_type=PolicyAssignmentObjectTypes.VIEW, object_statement=view_definition, input_string=current_instance_columns_string))
    else:
        current_instance_view_definition_without_columns_string = string_util.add_create_or_replace(replace_columns_string(object_statement=view_definition, old_columns_string=current_instance_columns_string, new_columns_string=''))

    current_instance_view_definition_with_columns_string = remove_database_name_from_view_statement(current_instance_view_definition_with_columns_string )
    current_instance_view_definition_without_columns_string = remove_database_name_from_view_statement(current_instance_view_definition_without_columns_string )

    return current_instance_view_definition, current_instance_view_definition_with_columns_string, current_instance_view_definition_without_columns_string

def add_policies_to_columns(object_type: PolicyAssignmentObjectTypes, object_statement: str, columns: list[ColumnInstance], policy_assignments_of_object: list[Dict]) -> Tuple[str, str]:
    """
    Add CMP assignments to the columns of a "table-like" object statement. Replaces existing columns string or adds a new columns string.
    Currently supported object types: VIEW, TABLE, DYNAMICTABLE
    """
    object_statement_with_policies_on_columns=object_statement
    
    if object_type==PolicyAssignmentObjectTypes.VIEW:
        object_statement_with_policies_on_columns_without_database_reference=object_statement
        columns_string = check_object_statement_for_columns_string(object_statement, columns)
            
        if columns_string:
            columns_string_with_policy_assignments, columns_string_with_policies_without_database_reference = add_policy_assignments_to_columns_string(columns_string=columns_string, columns=columns, policy_assignments_of_object=policy_assignments_of_object)
            if columns_string_with_policy_assignments:
                object_statement_with_policies_on_columns=replace_columns_string(object_statement, columns_string, columns_string_with_policy_assignments)
                object_statement_with_policies_on_columns_without_database_reference=replace_columns_string(object_statement, columns_string, columns_string_with_policies_without_database_reference)
        else:
            columns_string_, columns_string_without_database_reference_ =create_columns_string_of_view_with_policy_assignments(columns=columns, policy_assignments_of_object=policy_assignments_of_object)
            if columns_string_:
                object_statement_with_policies_on_columns=add_string_to_object_ddl(object_type, object_statement, columns_string_)
            if columns_string_without_database_reference_:
                object_statement_with_policies_on_columns_without_database_reference=add_string_to_object_ddl(object_type, object_statement, columns_string_without_database_reference_)

    elif object_type==PolicyAssignmentObjectTypes.DYNAMICTABLE:
        object_statement_with_policies_on_columns_without_database_reference = ''
        columns_string = check_object_statement_for_columns_string(object_statement, columns)
        
        if columns_string:
            columns_string_with_policy_assignments, _ = add_policy_assignments_to_columns_string(columns_string=columns_string, columns=columns, policy_assignments_of_object=policy_assignments_of_object)
            if columns_string_with_policy_assignments:
                object_statement_with_policies_on_columns = replace_columns_string(object_statement, columns_string, columns_string_with_policy_assignments)
        else:
            columns_string_ = create_columns_string_of_dynamictable_with_policy_assignments(columns=columns, policy_assignments_of_object=policy_assignments_of_object)
            if columns_string_:
                object_statement_with_policies_on_columns = add_string_to_object_ddl(object_type, object_statement, columns_string_)

    elif object_type==PolicyAssignmentObjectTypes.TABLE:
        object_statement_with_policies_on_columns_without_database_reference = ''
        columns_string = check_object_statement_for_columns_string(object_statement, columns)
        if not columns_string:
             raise ValueError('Columns definition not found in table DDL -> Something went wrong here!')
        else:
            columns_string_with_policy_assignments, _ = add_policy_assignments_to_columns_string(columns_string=columns_string, columns=columns, policy_assignments_of_object=policy_assignments_of_object)

            if columns_string_with_policy_assignments:
                object_statement_with_policies_on_columns=replace_columns_string(object_statement, columns_string, columns_string_with_policy_assignments)

    else:
        raise ValueError(f'Object type {object_type.value} not supported for policy assignment handling on columns!')
            
    return object_statement_with_policies_on_columns, object_statement_with_policies_on_columns_without_database_reference

def add_policy_assignments_to_columns_string(columns_string: str, columns: list[ColumnInstance], policy_assignments_of_object: list[Dict]) -> str:
    """
    Add CMP assignments to the columns definition of a table.
    """
    policy_assignments_on_columns=[policy_assignment for policy_assignment in policy_assignments_of_object if PolicyAssignmentType.get_policy_assignment_level(PolicyAssignmentType(policy_assignment["assignment_type"].upper()))==PolicyAssignmentLevel.TABLE_LEVEL_ASSIGNMENT]
    
    #sort columns by ordinal position
    columns=sorted(columns, key=lambda x: x.ordinal_position)

    if policy_assignments_on_columns:

        columns_string_without_sqlcomment = string_util.remove_comment(columns_string)
        columns_definitions=split_columns_string(columns_string_without_sqlcomment)

        #TODO write as function and align with create_columns_string_of_view_with_policy_assignments
        columns_string_with_policies='('
        columns_string_with_policies_without_database_reference='('
        for n_column, column in enumerate(columns):

            
            column_definition = columns_definitions[n_column]
            column_definition_without_database_reference = columns_definitions[n_column]
            
            # This regex_object defines all possible parameters that can be defined after the "WITH MASKING POLICY ..." part of the columns string
            regex_object=r"\s+with\s+projection\s+policy\s+|\s+projection\s+policy\s+|\s+with\s+tag\s*\(|\s+tag\s*\(|\s+comment\s*\'"
            column_definition_pre, column_definition_after = split_string_by_regex_match(regex_object, column_definition)
            
            
            column_name=get_column_name_from_column_definition(column_definition)
            if n_column+1 != column.ordinal_position:
                raise ValueError(f"While creating a columns-string for the view {column.object_schema}.{column.object_name} the ordinal position of the column {column.column_name} does not match the position in the columns string.")
            if column_name.upper() != column.column_name.upper():
                raise ValueError(f"The column name {column_name} from the column_definition does not match with the column name {column.column_name} from the list of column instances.")

            column_identifier= f"{column.object_schema}.{column.object_name}.{column.column_name}"

            policy_assignment_on_column=next((assignment for assignment in policy_assignments_on_columns if assignment["assignment"] == column_identifier), None)
            if policy_assignment_on_column:
                if policy_assignment_on_column["argument_columns"]:
                    using_string=f"{column.column_name}"
                    for arg_column in policy_assignment_on_column["argument_columns"]:
                        using_string = f"{using_string}, {arg_column}"
                    using_string = f"USING ({using_string})"
                else:
                    using_string=''
                column_definition= f'{column_definition_pre} WITH MASKING POLICY {policy_assignment_on_column["policy_database"]}.{policy_assignment_on_column["policy_schema"]}.{policy_assignment_on_column["policy"]} {using_string} {column_definition_after}, '
                column_definition_without_database_reference= f'{column_definition_pre} WITH MASKING POLICY {policy_assignment_on_column["policy_schema"]}.{policy_assignment_on_column["policy"]} {using_string}  {column_definition_after}, '            
            else:
                column_definition=f'{column_definition}, '
                column_definition_without_database_reference=f'{column_definition_without_database_reference}, '
            columns_string_with_policies=columns_string_with_policies + column_definition
            columns_string_with_policies_without_database_reference=columns_string_with_policies_without_database_reference + column_definition_without_database_reference

        columns_string_with_policies=columns_string_with_policies[:-2] + ')'
        columns_string_with_policies_without_database_reference=columns_string_with_policies_without_database_reference[:-2] + ')'

    else:
        columns_string_with_policies=''
        columns_string_with_policies_without_database_reference=''

    return columns_string_with_policies, columns_string_with_policies_without_database_reference

def get_column_name_from_column_definition(column_definition: str) -> str:
    """
    Returns the column name from a column definition.
    E.g. with column_definition = "column_1 INT" -> column_name = "column_1".
    """
    regex_column_name=r'^\s*(\S+)'
    match=re.match(regex_column_name, column_definition, flags=re.IGNORECASE)
    if match:
        column_name = match.group(1)
    return column_name

def split_columns_string(columns_string: str) -> list:
    """
    Splits a columns string by columns and ignores characters in single quotation marks and in brackets.

    Args:
        columns_string (str): The input columns string.

    Returns:
        list: List of column definitions.
    """
    columns=[]
    columns_string=columns_string[1:-1]
    stack = []
    ignore = False
    offset=0
    
    for i, char in enumerate(columns_string):
        if char == "'" and (i == 0 or columns_string[i-1] != "\\"):
            ignore = not ignore
        if ignore:
            continue
        if char == '(':
            stack.append('(')
        elif char == ')':
            stack.pop()
        if not stack and char == ',':
            end = i
            columns.append(columns_string[offset:end].strip())
            offset =  i+1
    if columns_string[offset:]:
        columns.append(columns_string[offset:].strip())
    return columns

def split_string_by_regex_match(regex_object: str, input_string: str) -> Tuple[str, str]:
    """
    Split string by regex match and ignores characters in single quotation marks and in brackets.
    Matches case-insensitive.
    """
    stack = []
    ignore = False
    ignore_i = 0

    substring_pre_match = input_string
    substring_after_match = ''
    
    for i, char in enumerate(input_string):
        if char == "'" and (i == 0 or input_string[i-1] != "\\"):
            if not ignore and not stack:
                string=input_string[ignore_i:i+1]
                match=re.search(regex_object, string , flags=re.IGNORECASE)
                if match:
                    substring_pre_match = input_string[:ignore_i+match.start()]
                    substring_after_match = input_string[ignore_i+match.start():]
                    break
            ignore = not ignore
            ignore_i = i
        if not ignore and char == ')':
            stack.pop()
            if not stack:
                ignore_i = i
        if not ignore and not stack:
            string=input_string[ignore_i:i+1]
            match=re.search(regex_object, string , flags=re.IGNORECASE)
            if match:
                substring_pre_match = input_string[:ignore_i+match.start()]
                substring_after_match = input_string[ignore_i+match.start():]
                break

        if not ignore and char == '(':
            stack.append('(')

    return substring_pre_match, substring_after_match

def add_column_with_policy(
        table_identifier: str, 
        column_name: str, 
        column_definition_without_comment: str, 
        column_comment_string: str, 
        policy_assignment_on_column: dict
        ) -> str:
    """
    Generate a statement to add a column with a policy assignment.
    """
    using_string = ""
    if policy_assignment_on_column['argument_columns']:
        using_string = f"{column_name}"
        for arg_column in policy_assignment_on_column['argument_columns']:
            using_string = f"{using_string}, {arg_column}"
        using_string = f"USING ({using_string})"

    add_column_with_policy_statement = f"ALTER TABLE {table_identifier} ADD COLUMN {column_definition_without_comment} WITH MASKING POLICY {policy_assignment_on_column['policy_database']}.{policy_assignment_on_column['policy_schema']}.{policy_assignment_on_column['policy']} {using_string} {column_comment_string}"
    return add_column_with_policy_statement

def add_policies_to_object(object_type: PolicyAssignmentObjectTypes, object_statement: str, columns: list[ColumnInstance], policy_assignments_of_object: list[Dict]) -> Tuple[str, str]:
    """
    Add an RAP assignment to a DDL statement of an Snowflake object (view or table).
    Currently supported object types: VIEW, TABLE, DYNAMICTABLE
    """
    object_statement_with_policies=object_statement
    object_statement_with_policies_without_database_reference=object_statement
    object_statement_with_policies_alternative=object_statement
    object_statement_with_policies_without_database_reference_alternative=object_statement

    if object_type==PolicyAssignmentObjectTypes.VIEW:

        columns_string = check_object_statement_for_columns_string(object_statement, columns)

        policy_assignment_on_objects=[policy_assignment for policy_assignment in policy_assignments_of_object if policy_assignment["assignment_type"]== 'views']
    
        if len(policy_assignment_on_objects) > 1:
            raise ValueError(f'There are multiple RAP assignments definied for the object {policy_assignment_on_objects[0]["assignment"]}.')
        elif len(policy_assignment_on_objects) == 1: 
            if not policy_assignment_on_objects[0]["argument_columns"]:
                raise ValueError(f'There are no assignment columns defined for the assignement of the RAP {policy_assignment_on_objects[0]["policy_database"]}.{policy_assignment_on_objects[0]["policy_schema"]}.{policy_assignment_on_objects[0]["policy"]} on the object {policy_assignment_on_objects["assignment"]}.')
            arg_columns_string=''
            for arg_column in policy_assignment_on_objects[0]["argument_columns"]:
                arg_columns_string = f"{arg_columns_string}, {arg_column}"
            arg_columns_string = f"ON ({arg_columns_string[2:]})"

            policy_string = f' WITH ROW ACCESS POLICY {policy_assignment_on_objects[0]["policy_database"]}.{policy_assignment_on_objects[0]["policy_schema"]}.{policy_assignment_on_objects[0]["policy"]} {arg_columns_string}'
            policy_string_without_database_reference = f' WITH ROW ACCESS POLICY {policy_assignment_on_objects[0]["policy_schema"]}.{policy_assignment_on_objects[0]["policy"]} {arg_columns_string}'

            object_statement_with_policies = add_string_to_object_ddl(object_type, object_statement, policy_string)
            object_statement_with_policies_without_database_reference = add_string_to_object_ddl(object_type, object_statement, policy_string_without_database_reference)

            #TODO write as function create_object_statement_with_policies_alternative and move to higher level
            if columns_string:
                if object_statement_with_policies.count(policy_string) ==1 and object_statement_with_policies.count(policy_string) == 1:
                    object_statement_with_policies_alternative=object_statement_with_policies.replace(policy_string, "{temp}").replace(columns_string, policy_string).replace("{temp}", columns_string)
                    object_statement_with_policies_without_database_reference_alternative=object_statement_with_policies_without_database_reference.replace(policy_string_without_database_reference, "{temp}").replace(columns_string, policy_string_without_database_reference).replace("{temp}", columns_string)
                else:
                    log.warning(f"Could not parse view definition such that the columns string is in front of the row access policy-assignment -> The policy assignments info might contain redundant information.")
                    object_statement_with_policies_alternative=object_statement_with_policies
                    object_statement_with_policies_without_database_reference_alternative=object_statement_with_policies_without_database_reference
            else:
                object_statement_with_policies_alternative=object_statement_with_policies
                object_statement_with_policies_without_database_reference_alternative=object_statement_with_policies_without_database_reference

    elif object_type==PolicyAssignmentObjectTypes.TABLE:
        policy_assignment_on_objects=[policy_assignment for policy_assignment in policy_assignments_of_object if policy_assignment["assignment_type"]== 'tables']

        #TODO write as function and align with object type 'VIEW' and 'DYNAMICTABLE'
        if len(policy_assignment_on_objects) > 1:
            raise ValueError(f'There are multiple RAP assignments definied for the object {policy_assignment_on_objects[0]["assignment"]}.')
        elif len(policy_assignment_on_objects) == 1: 
            if not policy_assignment_on_objects[0]["argument_columns"]:
                raise ValueError(f'There are no assignment columns defined for the assignement of the RAP {policy_assignment_on_objects[0]["policy_database"]}.{policy_assignment_on_objects[0]["policy_schema"]}.{policy_assignment_on_objects[0]["policy"]} on the object {policy_assignment_on_objects["assignment"]}.')
            arg_columns_string=''
            for arg_column in policy_assignment_on_objects[0]["argument_columns"]:
                arg_columns_string = f"{arg_columns_string}, {arg_column}"
            arg_columns_string = f"ON ({arg_columns_string[2:]})"

            policy_string = f' WITH ROW ACCESS POLICY {policy_assignment_on_objects[0]["policy_database"]}.{policy_assignment_on_objects[0]["policy_schema"]}.{policy_assignment_on_objects[0]["policy"]} {arg_columns_string}'

            object_statement_with_policies = add_string_to_object_ddl(object_type, object_statement, policy_string)

    elif object_type==PolicyAssignmentObjectTypes.DYNAMICTABLE:
        policy_assignment_on_objects=[policy_assignment for policy_assignment in policy_assignments_of_object if policy_assignment["assignment_type"]== 'dynamictables']

        #TODO write as function and align with object type 'VIEW' and 'TABLE'
        if len(policy_assignment_on_objects) > 1:
            raise ValueError(f'There are multiple RAP assignments definied for the object {policy_assignment_on_objects[0]["assignment"]}.')
        elif len(policy_assignment_on_objects) == 1: 
            if not policy_assignment_on_objects[0]["argument_columns"]:
                raise ValueError(f'There are no assignment columns defined for the assignement of the RAP {policy_assignment_on_objects[0]["policy_database"]}.{policy_assignment_on_objects[0]["policy_schema"]}.{policy_assignment_on_objects[0]["policy"]} on the object {policy_assignment_on_objects["assignment"]}.')
            arg_columns_string=''
            for arg_column in policy_assignment_on_objects[0]["argument_columns"]:
                arg_columns_string = f"{arg_columns_string}, {arg_column}"
            arg_columns_string = f"ON ({arg_columns_string[2:]})"

            policy_string = f' WITH ROW ACCESS POLICY {policy_assignment_on_objects[0]["policy_database"]}.{policy_assignment_on_objects[0]["policy_schema"]}.{policy_assignment_on_objects[0]["policy"]} {arg_columns_string}'

            object_statement_with_policies = add_string_to_object_ddl(object_type, object_statement, policy_string)

    else:
        raise ValueError(f'Object type {object_type.value} not supported for policy assignment handling on object!')
    
    return object_statement_with_policies, object_statement_with_policies_alternative, object_statement_with_policies_without_database_reference, object_statement_with_policies_without_database_reference_alternative

def replace_columns_string(object_statement: str, old_columns_string: str, new_columns_string: str) -> str:
    """
    Replace an columns_string with a new columns_string in the DDL of a Snowflake View.
    """
    object_statement = object_statement.replace(old_columns_string,new_columns_string,1)
    return object_statement

def add_string_to_object_ddl(object_type: PolicyAssignmentObjectTypes, object_statement: str, input_string: str) -> str:
    """
    Add a columns_string or an "with policy" statement to a DDL of a Snowflake object.
    Currently supported object types: VIEW, TABLE, DYNAMICTABLE
    """
    if object_type == PolicyAssignmentObjectTypes.VIEW:
        regex_object=fr"\s*CREATE\s+(?:OR\s+REPLACE\s+)?(?:SECURE\s+)?(?:RECURSIVE\s+)?VIEW\s+(?:IF\s+NOT\s+EXISTS\s+)?{string_util.REGEX_OBJ_IDENTIFIER_2}"
    elif object_type == PolicyAssignmentObjectTypes.TABLE:
        regex_object=fr"\s*CREATE\s+(?:OR\s+REPLACE\s+)?(?:TRANSIENT\s+)?TABLE\s+(?:IF\s+NOT\s+EXISTS\s+)?{string_util.REGEX_OBJ_IDENTIFIER_1}"
    elif object_type == PolicyAssignmentObjectTypes.DYNAMICTABLE:
        regex_object=fr"\s*CREATE\s+(?:OR\s+REPLACE\s+)?DYNAMIC\s+TABLE\s+(?:IF\s+NOT\s+EXISTS\s+)?{string_util.REGEX_OBJ_IDENTIFIER_1}"
    else:
        raise ValueError(f"Object type {object_type.value} not supported for add_string_to_object_ddl.")
    
    match=re.match(regex_object, object_statement, flags=re.IGNORECASE)

    if not match:
        raise ValueError(f"Object signature not found in DDL of object type {object_type.value}. Is this a valid DDL statement?\n{object_statement}")
    
    match=match[0]
    added_columns_string=f"{match} {input_string} "
    object_statement_with_columns_string, count=re.subn(
            regex_object,
            added_columns_string,
            object_statement,
            count=1,
            flags=re.IGNORECASE,
        )
    
    if count == 0:
        raise ValueError(
            f"FAILED to identify table signature in DDL. Is this a valid DDL statement?\n{object_statement}"
        )

    return object_statement_with_columns_string

def handle_policy_assignments_for_object_creation(object_type: PolicyAssignmentObjectTypes, object_statement: str, columns: list[ColumnInstance], policy_assignments_of_object: list[Dict]) -> list:
    """
    Handle the policy assignments for a new object creation based on predefined policy assignments.
    Adds the policy assignments directly to the DDL of the object.
    """
    statements=[f'USE SECONDARY ROLES {policy_assignments_of_object[0]["policy_pipeline_role"]};']
    
    object_statement, _ = add_policies_to_columns(
        object_type= object_type, 
        object_statement=object_statement, 
        columns=columns, 
        policy_assignments_of_object=policy_assignments_of_object
        )
    object_statement, _, _, _ = add_policies_to_object(
        object_type= object_type, 
        object_statement=object_statement, 
        columns=columns, 
        policy_assignments_of_object=policy_assignments_of_object
        )

    statements.append(object_statement)

    statements.append(f"USE SECONDARY ROLES NONE;")

    return statements

def handle_policy_assignments_for_alter_view_action(view_statement: str, columns: list[ColumnInstance], policy_assignments_info_from_object: dict) -> Tuple[list, str, str, str, str]:
    """
    Handle the policy assignments when altering a view.
    Adds the policy assignments directly to the DDL of the view.

    Background:
    Views are altered with CREATE OR REPLACE,
    and policy assignments need to be included in the CREATE OR REPLACE statement - otherwise they would be lost on re-creation!
    """

    statements=[f'USE SECONDARY ROLES {policy_assignments_info_from_object["assignments"][0]["policy_pipeline_role"]};']
    
    (
        view_statement_with_policies, 
        view_statement_with_policies_without_db_reference
    ) = add_policies_to_columns(
        object_type= PolicyAssignmentObjectTypes.VIEW,
        object_statement=view_statement,
        columns=columns,
        policy_assignments_of_object=policy_assignments_info_from_object["assignments"]
    )
    
    (
        _, 
        _, 
        view_statement_with_policies_without_db_reference, 
        view_statement_with_policies_without_db_reference_alternative
    )  = add_policies_to_object(
        object_type= PolicyAssignmentObjectTypes.VIEW,
        object_statement=view_statement_with_policies_without_db_reference,
        columns=columns,
        policy_assignments_of_object=policy_assignments_info_from_object["assignments"]
    )
    
    (
        view_statement_with_policies, 
        view_statement_with_policies_alternative, 
        _, 
        _
    ) = add_policies_to_object(
        object_type= PolicyAssignmentObjectTypes.VIEW,
        object_statement=view_statement_with_policies,
        columns=columns,
        policy_assignments_of_object=policy_assignments_info_from_object["assignments"]
    )

    statements.append(view_statement_with_policies)

    statements.append(f"USE SECONDARY ROLES NONE;")

    return statements, view_statement_with_policies, view_statement_with_policies_alternative, view_statement_with_policies_without_db_reference, view_statement_with_policies_without_db_reference_alternative

def handle_policy_assignments_for_alter_table_add_column_action(table_identifier: str, column: str, column_comment: str, column_definition: str, policy_assignments_info_from_object: dict= {}):
    """
    Handle the policy assignments when altering a table such that a column is added.
    Includes the policy assignments directly in the ADD COULMN statement.
    """
    statements=[]
    if column_comment:
        regex_column_comment= f"\s*comment\s*'{re.escape(column_comment)}'"
        column_definition_escaped = column_definition.replace('\\','')
        match=re.search(regex_column_comment, column_definition, flags=re.IGNORECASE)
        if not match:
            match_=re.search(regex_column_comment, column_definition_escaped, flags=re.IGNORECASE)
    else:
        match=None
        match_=None

    if match:
        column_comment_string = match.group(0)
        column_definition_without_comment = column_definition.replace(column_comment_string, '')
    elif match_:
        column_comment_string = match_.group(0)
        column_definition_without_comment = column_definition_escaped.replace(column_comment_string, '')
    else:
        column_comment_string = ''
        column_definition_without_comment = column_definition

    policy_assignment_on_column= next(
        (assignment for assignment in policy_assignments_info_from_object['assignments'] 
        if assignment['policy_type']=='column_masking_policies' 
        and assignment['assignment'].split('.')[2].upper()==column.upper()),
        None
        )

    if policy_assignment_on_column:
        
        statements.append(f'USE SECONDARY ROLES {policy_assignment_on_column["policy_pipeline_role"]}')

        statements.append(
            add_column_with_policy(
                table_identifier=table_identifier,
                column_name = column,
                column_definition_without_comment=column_definition_without_comment,
                column_comment_string=column_comment_string,
                policy_assignment_on_column=policy_assignment_on_column,
            )
        )
        statements.append(f"USE SECONDARY ROLES NONE")
    else:
        statements.append(
                    f"ALTER TABLE {table_identifier} ADD COLUMN {column_definition}"
                )
    
    return statements

class PolicyService:
    """
    Class to handle the interaction with SNOWFLAKE POLICIES.
    An instance of this class is scoped to a single database (defined by the given snow_client).
    """

    def __init__(
        self,
        policy_assignments_handler: Dict,
        policy_assignments_target_database: str,
        policy_assignments_config_file_path: str,
        policy_assignments_project: str,
        policy_assignments_info_output_folder_path: str,
        policy_assignments_repo_path: str = ''
    ):
        """
        Initializes a policy service.
        """
        self.policy_assignments_handler = policy_assignments_handler
        self.policy_assignments_info = {}
        self.policy_assignments_target_database = policy_assignments_target_database
        self.policy_assignments_config_file_path = policy_assignments_config_file_path
        self.policy_assignments_project = policy_assignments_project
        self.policy_assignments_info_output_folder_path = policy_assignments_info_output_folder_path
        self.policy_assignments_repo_path = policy_assignments_repo_path
        self.module_root_folder = os.path.abspath(os.path.join(os.path.dirname(__file__), "..", ".."))
        self.project_config = self._get_project_config()
        self._check_target_db_in_policy_config()
        self.load_devops_variables()

    def load_devops_variables(self):
        self.build_reason = os.environ.get("BUILD_REASON")
        self.build_repository_name = os.environ.get("BUILD_REPOSITORY_NAME")
        self.system_accesstoken = os.environ.get("SYSTEM_ACCESSTOKEN")
        self.system_teamprojectid = os.environ.get("SYSTEM_TEAMPROJECTID")
        self.system_teamfoundationcollectionuri = os.environ.get("SYSTEM_TEAMFOUNDATIONCOLLECTIONURI")
        self.system_pullrequest_pullrequestid = os.environ.get("SYSTEM_PULLREQUEST_PULLREQUESTID")

    @staticmethod
    def parse_instance_for_policy_assignments_handler(instance_object: Union[InstanceObject,None]) -> Union[str,Dict]:
        """
        Parse instance object for policy assignments handler.
        """
        if not instance_object:
            return "", {}
        else:
            if 'table_columns' in instance_object.__dict__.keys():
                columns=[f"{instance_object.full_name}.{column.column_name}" for column in instance_object.table_columns]
            else:
                columns=[]

            return instance_object.full_name, columns
         
    def get_policy_assignments_info(self, get_all_assignments: bool = False) -> Dict:
        """
        Get the policy assignments configured in the SNOWFLAKE POLICIES repo for all objects in the policy assignment handler.
        """
        self._check_target_db_in_policy_config()

        self.project_config = self._get_project_config()

        policy_assignments_file_paths=self.project_config["POLICY_ASSIGNMENTS_FILES"]

        self._load_policy_assignments(policy_assignments_file_paths, get_all_assignments = get_all_assignments)

        self.policy_assignments_info = self.policy_assignments_handler
    
    def create_policy_assignments_review_page(self):
        """
        Create policy assignments review page for the pipeline log.
        """
        # TODO
        log.info()

    def save_policy_assignments_info(self):
        """
        Save policy assignments information to output folder.
        """
        timestamp = datetime.datetime.now().strftime("%Y-%m-%d_%H-%M_%S")
        if self.policy_assignments_info_output_folder_path:
            os.makedirs(self.policy_assignments_info_output_folder_path, exist_ok=True)
            filename = os.path.join(self.policy_assignments_info_output_folder_path, f"policy_assignments_info.{timestamp}.json")
            log.info(f"SAVING policy assignments info as JSON file in '{filename}'")
            with open(filename, 'w') as f:
                json.dump(self.policy_assignments_info, f, indent=4)

    def _create_pr_comment(self, comment_markdown: str):
        """
        Create a comment in an Azure DevOps Pull Request.
        """
        if self.system_accesstoken:
            authorization = str(
                base64.b64encode(bytes(":" + self.system_accesstoken, "ascii")),
                "ascii",
            )
        else:
            raise ValueError(f"System Accesstoken for Azure DevOps not found while trying to create a comment in a Pull Request.")

        headers = {
            "Content-Type": "application/json",
            "Authorization": "Basic " + authorization,
        }

        data = {
            "comments": [
                {"parentCommentId": 0, "content": comment_markdown, "commentType": 1}
            ],
            "status": 1,
        }

        url = f"{self.system_teamfoundationcollectionuri}{self.system_teamprojectid}/_apis/git/repositories/{self.build_repository_name}/pullRequests/{self.system_pullrequest_pullrequestid}/threads?api-version=7.0"

        response = requests.post(url=url, headers=headers, verify=True, json=data)

        if response.status_code == 200:
            api_response = json.loads(response.text)
            return api_response, True
        else:
            log.info(f"request url: {url}")
            log.info(f"request response.status_code: {response.status_code}")
            log.info(f"request response.text: {response.text}")
            raise EnvironmentError(
                f"Error while trying to create a comment on pull request {self.system_pullrequest_pullrequestid} - Pull Request Threads Rest API!"
            )

    def _check_target_db_in_policy_config(self):
        """
        Check if the target database is configured in the policy assignments config file.
        """
        if self.policy_assignments_target_database not in self.project_config["ENVIRONMENTS"].values():
            raise ValueError(f"Error while initializing the policy service. \
                            The target database {self.policy_assignments_target_database} is not defined for project {self.policy_assignments_project} in the config file {self.policy_assignments_config_file_path} !" 
                            )
        else:
            return True
                
    def _get_project_config(self):
        """
        Get list of the policy assignments JSON file paths from the config file.
        """
        config_dict = load_json(self.policy_assignments_config_file_path)
        if self.policy_assignments_project not in config_dict["PROJECT_CONFIGS"]:
            raise ValueError(f"Error while initializing the policy service. \
                             The project name {self.policy_assignments_project} is not defined in the config file {self.policy_assignments_config_file_path} !"
                             )
        else:
            project_config=config_dict["PROJECT_CONFIGS"][self.policy_assignments_project]
        return project_config
    
    def _load_policy_assignments(self, policy_assignments_file_paths: str, get_all_assignments: bool = False) -> None:
        """
        Load the policy assignments from JSON files for all objects in the policy_assignments_handler.
        """
        for policy_assignments_relative_file_path in policy_assignments_file_paths:
            
            if self.policy_assignments_repo_path:
                policy_assignments_absolute_file_path = PurePath(self.policy_assignments_repo_path).joinpath(policy_assignments_relative_file_path)

            else:
                policy_assignments_absolute_file_path = policy_assignments_relative_file_path

            if not os.path.isfile(policy_assignments_absolute_file_path):
                raise EnvironmentError(
                    f"Policy assignments file path [ '{policy_assignments_absolute_file_path}' ] is not valid."
                )

            log.debug(f"------------- LOAD policy assignments from [ '{policy_assignments_absolute_file_path}' ]")
            
            
            policy_assignments = load_json(policy_assignments_absolute_file_path)
            policy_assignments.pop("$schema")

            log.debug(f"------------- Policy assignments information from [ '{policy_assignments_absolute_file_path}' ]: {policy_assignments}\n") 

            for policy_type in ["column_masking_policies", "row_access_policies"]:
                for policy, assignments_groups in policy_assignments[policy_type].items():

                    for assignment_type, assignments_group in assignments_groups.items():

                        if assignment_type == "annotations":
                            if "person_responsible" in assignments_group:
                                person_responsible = assignments_group["person_responsible"]
                            else:
                                person_responsible = ''
                            continue

                        object_type=PolicyAssignmentType.get_object_type_from_assignment_type(assignment_type).value

                        if object_type not in self.policy_assignments_handler:
                            continue

                        if assignment_type.upper() != PolicyAssignmentType.TAGS.value or get_all_assignments:
                            for assignment, argument_columns in assignments_group.items():

                                for instance in ["current_instance", "desired_instance"]:
                                    
                                    if get_all_assignments:
                                        if policy_type == "column_masking_policies":
                                            object_identifier = assignment.rsplit('.',1)[0].upper()
                                        else:
                                            object_identifier = assignment.upper()
                                        assignment_info = {
                                                "policy_type": policy_type, 
                                                "assignment_type":assignment_type, 
                                                "policy": policy, 
                                                "assignment":assignment,
                                                "argument_columns": argument_columns,
                                                "policy_pipeline_role": self.project_config["SNOWFLAKE_CREDENTIALS"]["ROLE"], 
                                                "policy_schema":self.project_config["POLICY_SCHEMA"], 
                                                "policy_database":self.policy_assignments_target_database,
                                                "person_responsible": person_responsible}
                                        
                                        if not object_identifier in self.policy_assignments_handler[object_type]:
                                            self.policy_assignments_handler[object_type][object_identifier]={"policy_handling_type": "NEW_OBJECT", "current_instance":{"assignments":[]},"desired_instance":{"assignments":[]}}

                                        self.policy_assignments_handler[object_type][object_identifier][instance]["assignments"].append(assignment_info)

                                    elif PolicyAssignmentType.get_policy_assignment_level(PolicyAssignmentType(assignment_type.upper())) == PolicyAssignmentLevel.TABLE_LEVEL_ASSIGNMENT:

                                        object_name = assignment.rsplit('.',1)[0].upper()

                                        if (object_name in self.policy_assignments_handler[object_type]
                                            and instance in self.policy_assignments_handler[object_type][object_name]
                                            and assignment in self.policy_assignments_handler[object_type][object_name][instance]["columns"]):

                                            if "assignments" not in self.policy_assignments_handler[object_type][object_name][instance]:
                                                self.policy_assignments_handler[object_type][object_name][instance]["assignments"]=[]

                                            self.policy_assignments_handler[object_type][object_name][instance]["assignments"].append({
                                                "policy_type": policy_type, 
                                                "assignment_type":assignment_type, 
                                                "policy": policy, 
                                                "assignment":assignment,
                                                "argument_columns": argument_columns,
                                                "policy_pipeline_role": self.project_config["SNOWFLAKE_CREDENTIALS"]["ROLE"], 
                                                "policy_schema":self.project_config["POLICY_SCHEMA"], 
                                                "policy_database":self.policy_assignments_target_database,
                                                "person_responsible": person_responsible})
                                            
                                    else:
                                        object_name = assignment
                                        if (object_name in self.policy_assignments_handler[object_type]
                                            and instance in self.policy_assignments_handler[object_type][object_name]):

                                            if "assignments" not in self.policy_assignments_handler[object_type][object_name][instance]:
                                                self.policy_assignments_handler[object_type][object_name][instance]["assignments"]=[]

                                            self.policy_assignments_handler[object_type][object_name][instance]["assignments"].append({
                                                "policy_type": policy_type, 
                                                "assignment_type":assignment_type, 
                                                "policy": policy, 
                                                "assignment":assignment, 
                                                "argument_columns": argument_columns,
                                                "policy_pipeline_role": self.project_config["SNOWFLAKE_CREDENTIALS"]["ROLE"], 
                                                "policy_schema":self.project_config["POLICY_SCHEMA"], 
                                                "policy_database":self.policy_assignments_target_database,
                                                "person_responsible": person_responsible})
                                            
                        elif assignment_type.upper() == PolicyAssignmentType.TAGS.value:
                            for assignment in assignments_group:
                                for instance in ["current_instance", "desired_instance"]:
                                    object_name = assignment
                                    if (object_name in self.policy_assignments_handler[object_type]
                                        and instance in self.policy_assignments_handler[object_type][object_name]):

                                        if "assignments" not in self.policy_assignments_handler[object_type][object_name][instance]:
                                            self.policy_assignments_handler[object_type][object_name][instance]["assignments"]=[]

                                        self.policy_assignments_handler[object_type][object_name][instance]["assignments"].append({
                                            "policy_type": policy_type, 
                                            "assignment_type":assignment_type, 
                                            "policy": policy, 
                                            "assignment":assignment, 
                                            "policy_pipeline_role": self.project_config["SNOWFLAKE_CREDENTIALS"]["ROLE"], 
                                            "policy_schema":self.project_config["POLICY_SCHEMA"], 
                                            "policy_database":self.policy_assignments_target_database,
                                            "person_responsible": person_responsible})
                                        
        log.debug(f"------------- Policy assignments information for handler objects:\n {self.policy_assignments_handler}")                            

    def create_azure_devops_pr_comment_from_policy_assignments_info(self):
        """
        Create a comment a pull request in Azure DevOps containing the policy assignments info.
        """
        if self.build_reason == "PullRequest":
            comment_markdown_policy_assignments_info = self._parse_policy_assignments_info()
            if comment_markdown_policy_assignments_info:
                self._create_pr_comment(comment_markdown_policy_assignments_info)
                             
    def _parse_policy_assignments_info(
        self
    ) -> str:
        """
        Parse the policy assignments info in the format of a markdown table.
        """
        comment_markdown_policy_assignments_info = ""

        if self.policy_assignments_info:

            comment_markdown_policy_assignments_info = f"""
            The following policy assignments are handled during the deployment on target database {self.policy_assignments_target_database} in projetct {self.policy_assignments_project} 
            (Note: This is a summary. Please see the pipeline artifacts for the full information in JSON format!): \n"""

            for object_type in self.policy_assignments_info.keys():

                for object_identifier, policy_handling in self.policy_assignments_info[object_type].items():

                    for instance in ['current_instance', 'desired_instance']:

                        table = "| Policy Type | Policy | Assignment |\n"
                        table += "| --- | --- | --- |\n"

                        if instance in policy_handling and "assignments" in policy_handling[instance]:
                            for assignment in policy_handling[instance]["assignments"]:
                                row = f"| {assignment['policy_type']} | {assignment['policy']} | {assignment['assignment']} |\n"
                                table += row

                        if instance == 'current_instance':
                            table_current_assignments=table
                        else:
                            table_desired_assignments=table

                    comment_markdown_policy_assignments_info += (
                        f"##  {policy_handling['policy_handling_type']} {object_type} {object_identifier}\n\n ### current assignments \n {table_current_assignments}\n\n ### future assignments \n {table_desired_assignments}"
                    )

        return comment_markdown_policy_assignments_info
        



import re
from enum import Enum
from typing import Dict, List

import git

import aceutils.file_util as file_util
import aceutils.string_util as string_util


class PreOrPostDeploymentScriptsExecutionOptions(Enum):
    """
    Available option for the execution of pre/postdeployment scripts.
    Applies to all scripts.
    """
    ALL = "all"
    GIT = "git"
    NONE = "none"


class PreOrPostDeploymentScriptTarget(Enum):
    TARGET = "targetDatabase"
    META = "metaDatabase"


class PreOrPostDeploymentScriptType(Enum):
    """
    Enum for valid types of post deployment scripts
    """

    PYTHON = "PYTHON"
    SQL = "SQL"

    @classmethod
    def get_type_from_path(cls, script_path: str) -> "PreOrPostDeploymentScriptType":
        """
            Converts a file path based on file extension to a DeploymentScriptType instance
        Args:
            script_type: str - absolute script path to file
        Raises:
            ValueError - If file path is None or empty
            NotImplementedError - If file extension is not supported
        Returns:
            Instance of PreOrPostDeploymentScriptType
        """
        if script_path is None or script_path == "":
            raise ValueError("script_path should not be empty")

        if script_path.lower().endswith(".sql"):
            return PreOrPostDeploymentScriptType.SQL
        if script_path.lower().endswith(".py"):
            return PreOrPostDeploymentScriptType.PYTHON
        raise NotImplementedError(f"script type not supported for '{script_path}'.")


class PreOrPostDeploymenStep(object):
    """
    Descripts a post deployment step as part of the acedeploy deployment
    """

    def __init__(
        self,
        path: str,
        git_change_type: git.DiffIndex.change_type = None,
        string_replace_dict: Dict[str, str] = None,
        regex_filter_list: List[str] = None,
    ):
        """
            Inits a new instance
        Args:
            path: str - absolute path to file
            git_change_type: git.DiffIndex.change_type - type of git change associated to this file (None if no change or unknown)
            string_replace_dict: Dict[str,str] - optional, contains information about strings which need to be replaced in SQL DDLs
            regex_filter_list: List[str] - optional, contains a list of regular expressions (case insensitive). If any regex matches against the content, an exception will be raised
        """
        if string_replace_dict is None:
            string_replace_dict = {}
        if regex_filter_list is None:
            regex_filter_list = []
        self.path = path
        self.git_change_type = git_change_type
        self.type = PreOrPostDeploymentScriptType.get_type_from_path(self.path)
        if self.type == PreOrPostDeploymentScriptType.PYTHON:
            raise Exception("Python pre/postdeployment files are not allowed")
        content = self._load_sql_file(path)
        self.content = string_util.string_replace_by_dict(content, string_replace_dict)
        unapproved_command = self._contains_forbidden_command(
            self.content, regex_filter_list
        )
        if unapproved_command:
            raise Exception(
                f"Pre/Postdeployment script [ '{self.path}' ] contains forbidden statement [ '{unapproved_command}' ]."
            )
        self.execute_step = None  # boolean: can be set later to indicate if this step should be run during deployment
        self.target = PreOrPostDeploymentScriptTarget.TARGET # PreOrPostDeploymentScriptTarget: inidicates if this step should be run on the target database or the meta database


    def set_target(self, target: PreOrPostDeploymentScriptTarget) -> None:
        self.target = target


    def set_execute_step(self, pre_and_postdeployment_execution: PreOrPostDeploymentScriptsExecutionOptions, step_condition: str) -> None:
        if pre_and_postdeployment_execution == PreOrPostDeploymentScriptsExecutionOptions.NONE:
            self.execute_step = False
        elif pre_and_postdeployment_execution == PreOrPostDeploymentScriptsExecutionOptions.ALL:
            if step_condition == "onChange":
                raise ValueError("It is not allowed to use 'onChange' when 'all' pre/postdeployment scripts should be executed (ignoreGitInformation = True)")
            elif step_condition == "never":
                self.execute_step = False
            else:
                self.execute_step = True
        elif pre_and_postdeployment_execution == PreOrPostDeploymentScriptsExecutionOptions.GIT:
            if step_condition == "always":
                self.execute_step = True
            elif step_condition == "onChange":
                self.execute_step = self.git_change_type in ("A", "R", "M")
            elif step_condition == "never":
                self.execute_step = False
            else:
                raise ValueError("step_condition must be one of 'always', 'onChange', 'never'.")


    @staticmethod
    def _load_sql_file(file_path: str) -> str:
        """
        Load a sql file from a given file path.
        """
        content = file_util.load(file_path)
        return string_util.remove_comment_after_statement(content)

    @staticmethod
    def _contains_forbidden_command(content: str, filter_list: List[str]) -> str:
        """
        Check if given sql file content matches against one of the regular expressions given in filter_list (case insensitive).
        If yes, return that command.
        """
        cleaned_content = string_util.remove_text_in_quotes(
            string_util.remove_comment(content)
        )
        for forbidden_words in filter_list:
            result = re.search(forbidden_words, cleaned_content, re.IGNORECASE)
            if result:
                return result.group()
        return False



import logging
from typing import List

from aceaccount.core.model_account_object_action_entities import AccountObjectAction
from aceaccount.core.model_account_object_sql_entities import AccountObjectActionType
from aceaccount.services.account_objects_metadata_service import AccountMetadataService
from aceaccount.services.account_objects_solution_service import AccountSolutionClient
from aceservices.snowflake_service import SnowClientConfig
from aceutils.logger import LoggingAdapter

logger = logging.getLogger(__name__)
log = LoggingAdapter(logger)


class AccountObjectCompareClient(object):
    """
    Compare the state of the account objects (current state) to the input file (desired state).
    Generate a list of actions to change current state into desired state.
    """

    def __init__(
        self,
        desired_state: AccountSolutionClient,
        current_state: AccountMetadataService,
        snow_client_config: SnowClientConfig
    ) -> None:
        """
            Inits a new DbCompareClient
        Args:
            desired_state: AccountMetadataService - State that should be achieved as defined by the input file
            current_state: MetadataService - Current account state
        """
        self._desired_state = desired_state
        self._current_state = current_state
        self._snow_client_config = snow_client_config
        self._action_list = []
        self._grant_action_list = []
        self._tag_action_list = []

    @property
    def action_list(self) -> List[AccountObjectAction]:
        return self._action_list
    
    @property
    def grant_action_list(self) -> List[AccountObjectAction]:
        return self._grant_action_list
    
    @property
    def tag_action_list(self) -> List[AccountObjectAction]:
        return self._tag_action_list

    def generate_add_actions(self) -> None:
        """
        Get all ADD actions and add them to action list.
        Add all objects that are in _desired_state, but not in _current_state.
        """
        log.debug("ADD actions of type [ 'ADD' ] to action_list")
        for desired_obj in self._desired_state.all_account_objects:
            current_obj = self._current_state.get_object_by_object(desired_obj)
            if not current_obj:
                log.debug(
                    f"ADD action of type [ 'ADD' ] for object [ '{str(desired_obj)}' ]"
                )
                self.action_list.append(
                    AccountObjectAction.factory(
                        name=desired_obj.name,
                        object_type=desired_obj.object_type,
                        action_type=AccountObjectActionType.ADD,
                        snow_client_config=self._snow_client_config,
                        desired_instance=desired_obj,
                    )
                )

    def generate_drop_actions(self, drop_enabled: dict) -> None:
        """
        Get all ADD actions and add them to action list.
        Add all objects that are in _current_state, but not in _desired_state.
        """
        log.debug("ADD actions of type [ 'DROP' ] to action_list")
        for current_obj in self._current_state.all_account_objects:
            desired_obj = self._desired_state.get_object_by_object(current_obj)

            if current_obj.object_type.value not in drop_enabled:
                raise ValueError(f"No dropEnabled parameter-configuration for the AccountObjectType {current_obj.object_type.value} found. Please adjust the config.json !")
            
            elif not desired_obj and drop_enabled[current_obj.object_type.value]:
                log.debug(
                    f"ADD action of type [ 'DROP' ] for object [ '{str(current_obj)}' ]"
                )
                self.action_list.append(
                    AccountObjectAction.factory(
                        name=current_obj.name,
                        object_type=current_obj.object_type,
                        action_type=AccountObjectActionType.DROP,
                        snow_client_config=self._snow_client_config,
                        current_instance=current_obj,
                    )
                )

    def generate_alter_actions(self) -> None:
        """
        Get all ADD actions and add them to action list.
        Add all objects that have different states in _current_state and in _desired_state.
        """
        log.debug("ADD actions of type [ 'ALTER' ] to action_list")
        for current_obj in self._current_state.all_account_objects:
            desired_obj = self._desired_state.get_object_by_object(current_obj)
            if (desired_obj is not None) and (desired_obj != current_obj):
                log.debug(
                    f"ADD action of type [ 'ALTER' ] for object [ '{str(current_obj)}' ]"
                )
                self.action_list.append(
                    AccountObjectAction.factory(
                        name=current_obj.name,
                        object_type=current_obj.object_type,
                        action_type=AccountObjectActionType.ALTER,
                        snow_client_config=self._snow_client_config,
                        current_instance=current_obj,
                        desired_instance=desired_obj,
                    )
                )

    def generate_grant_actions(self) -> None:
        """
        Get all GRANT actions and add them to action list.
        Add all grants that are in _desired_state, but not in _current_state.
        """
        log.debug("ADD actions of type [ 'GRANT' ] to action_list")
        for current_obj in self._current_state.all_account_objects:
            desired_obj = self._desired_state.get_object_by_object(current_obj)
            if (desired_obj is not None) and  hasattr(desired_obj, "grants") and desired_obj.grants:
                if  (not hasattr(current_obj, "grants") or ({key:set(value) for key,value in desired_obj.grants.items()} != {key:set(value) for key,value in current_obj.grants.items()})):
                    log.debug(
                        f"ADD action of type [ 'GRANT' ] for object [ '{str(current_obj)}' ]"
                    )
                    self.grant_action_list.append(
                        AccountObjectAction.factory(
                            name=current_obj.name,
                            object_type=current_obj.object_type,
                            action_type=AccountObjectActionType.GRANT,
                            snow_client_config=self._snow_client_config,
                            current_instance=current_obj,
                            desired_instance=desired_obj,
                        )
                    )

    def generate_revoke_actions(self) -> None:
        """
        Get all REVOKE actions and add them to action list.
        Add all grants that are in _current_state, but not in _desired_state.
        """
        log.debug("ADD actions of type [ 'REVOKE' ] to action_list")
        for current_obj in self._current_state.all_account_objects:
            desired_obj = self._desired_state.get_object_by_object(current_obj)
            if (desired_obj is not None) and hasattr(current_obj, "grants") and current_obj.grants:
                if hasattr(desired_obj, "grants") and ({key:set(value) for key,value in desired_obj.grants.items()} != {key:set(value) for key,value in current_obj.grants.items()}):
                    log.debug(
                        f"ADD action of type [ 'REVOKE' ] for object [ '{str(current_obj)}' ]"
                    )
                    self.grant_action_list.append(
                        AccountObjectAction.factory(
                            name=current_obj.name,
                            object_type=current_obj.object_type,
                            action_type=AccountObjectActionType.REVOKE,
                            snow_client_config=self._snow_client_config,
                            current_instance=current_obj,
                            desired_instance=desired_obj,
                        )
                    )

    def generate_settag_actions(self) -> None:
        """
        Get all SETTAG actions and add them to action list.
        Add all tags that are in _desired_state, but not in _current_state.
        """
        log.debug("ADD actions of type [ 'SETTAG' ] to action_list")
        for current_obj in self._current_state.all_account_objects:
            desired_obj = self._desired_state.get_object_by_object(current_obj)
            if (desired_obj is not None) and hasattr(desired_obj, "tags") and desired_obj.tags:
                if (not hasattr(current_obj, "tags") or ({key.upper():value for key,value in desired_obj.tags.items()} != {key.upper():value for key,value in current_obj.tags.items()})):
                    log.debug(
                        f"ADD action of type [ 'SETTAG' ] for object [ '{str(current_obj)}' ]"
                    )
                    self.tag_action_list.append(
                        AccountObjectAction.factory(
                            name=current_obj.name,
                            object_type=current_obj.object_type,
                            action_type=AccountObjectActionType.SETTAG,
                            snow_client_config=self._snow_client_config,
                            current_instance=current_obj,
                            desired_instance=desired_obj,
                        )
                    )

    def generate_unsettag_actions(self) -> None:
        """
        Get all UNSETTAG actions and add them to action list.
        Add all tags that are in _current_state, but not in _desired_state.
        """
        log.debug("ADD actions of type [ 'UNSETTAG' ] to action_list")
        for current_obj in self._current_state.all_account_objects:
            desired_obj = self._desired_state.get_object_by_object(current_obj)
            if (desired_obj is not None) and hasattr(current_obj, "tags") and current_obj.tags:
                if hasattr(desired_obj, "tags") and ({key.upper():value for key,value in desired_obj.tags.items()} != {key.upper():value for key,value in current_obj.tags.items()}):
                    log.debug(
                        f"ADD action of type [ 'UNSETTAG' ] for object [ '{str(current_obj)}' ]"
                    )
                    self.tag_action_list.append(
                        AccountObjectAction.factory(
                            name=current_obj.name,
                            object_type=current_obj.object_type,
                            action_type=AccountObjectActionType.UNSETTAG,
                            snow_client_config=self._snow_client_config,
                            current_instance=current_obj,
                            desired_instance=desired_obj,
                        )
                    )


from abc import ABC, abstractmethod

from aceaccount.core.model_account_object_sql_entities import AccountObjectType

class AccountObject(ABC):
    def __init__(self, name: str, object_type: AccountObjectType):
        self.name = name
        self.object_type = object_type

    @abstractmethod
    def __str__(self):
        pass

    @property
    def id(self):
        return f"{str(self.object_type)} {self.name.upper()}"

    @property
    def name(self):
        return self._name.replace('"', "").upper()

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def full_name(self):
        return f"{self.name}"

    @property
    def object_type(self):
        return self._object_type

    @object_type.setter
    def object_type(self, value):
        if not isinstance(value, AccountObjectType):
            raise TypeError("Incorrect datatype. Allowed: AccountObjectType")
        self._object_type = value

    def compare_full_name(self, full_name_to_compare):
        """
        Test if name_to_compare matches the full name of this AccountObject.

        Returns true if full names match, else false.
        Ignores double quotes around the names.
        """
        return self.full_name == full_name_to_compare.upper().replace('"', "")

    def compare_id(self, id_to_compare):
        """
        Test if id_to_compare matches the id of this AccountObject.

        Returns true if names match, else false.
        Ignores double quotes around the names.
        """
        return self.id == id_to_compare.replace('"', "")



import difflib
import logging
import re
from acedeploy.services.dependency_parser import DependencyParser

import aceutils.database_management as database_management
import aceutils.string_util as string_util
from acedeploy.core.model_configuration import SolutionConfig
from acedeploy.core.model_instance_objects import InstanceTable
from acedeploy.core.model_sql_entities import DbObjectType
from acedeploy.main import deploy_to_meta_database
from acedeploy.services.metadata_service import MetadataService
from aceservices.snowflake_service import SnowClient, SnowClientConfig
from acedeploy.services.solution_service import SolutionClient
from aceutils.logger import LoggingAdapter

logger = logging.getLogger(__name__)
log = LoggingAdapter(logger)


def configure() -> SolutionConfig:
    """
        Loads the current configuration from the deployment.json file.
    Returns:
        SolutionConfigVCollection
    """
    return SolutionConfig()


def compare_files_and_database(
    config: SolutionConfig,
    detailed_output: bool = False,
    ignore_column_order: bool = True,
) -> None:
    """
        Compare solution to database
    Args:
        config: SolutionConfig - deployment configuration
        detailed_output: bool - print diff between both DDLs if no match (default: false)
        ignore_column_order: bool - ignore order of columns when comparing tables
    Returns:
        None
    """
    snow_client_account = SnowClient(SnowClientConfig.get_from_solution_config(config))
    database_management.create_database(
        config.database_meta,
        snow_client_account,
        replace_if_exists=True,
        drop_schema_public=True,
        db_retention_time_in_days=config.target_db_retention_time,
    )
    snow_client_meta = SnowClient(
        SnowClientConfig.get_from_solution_config(config, config.database_meta)
    )
    snow_client_target = SnowClient(
        SnowClientConfig.get_from_solution_config(config, config.database_target)
    )
    solution_client = SolutionClient(
        project_folder=config.project_folder,
        predeployment_folders=config.pre_deployment_folders,
        postdeployment_folders=config.post_deployment_folders,
        predeployment_settings=config.pre_deployment_settings,
        postdeployment_settings=config.post_deployment_settings,
        prepostdeployment_filter_list=config.prepostdeployment_filter_list,
        config_schema_list=config.schema_list,
        disabled_object_types=config.disabled_object_types,
    )
    solution_client.load_solution(string_replace_dict=config.string_replace)

    try:

        try:
            dependency_client = DependencyParser(solution_client)
            dependency_client.build_full_dependency_graph()
            log.info(f"DEPLOY full solution TO meta db [ '{config.database_meta}' ]")
            deploy_to_meta_database(
                solution_client=solution_client,
                dependency_client=dependency_client,
                snow_client=snow_client_meta,
                config=config,
                mode="solution",
                parallel_threads=config.parallel_threads,
            )
        except Exception as err:
            log.error("DEPLOYMENT to meta db FAILED. Solution might not be valid.")
            raise err

        missing_objects = []
        changed_objects = []

        # compare tables
        log.info(
            f"COMPARE METADATA (tables and views) ON target db [ '{config.database_target}' ] AND meta db [ '{config.database_meta}' ]"
        )
        (
            missing_tables,
            changed_tables,
            _,
            changed_views,
        ) = _compare_tables_and_views_metadata(
            snow_client_meta, snow_client_target, config, ignore_column_order
        )
        missing_objects.extend(missing_tables)
        changed_objects.extend([f"{o} (metadata mismatch)" for o in changed_tables])
        # missing_objects.extend(missing_views) # views will be added through ddl comparison
        changed_objects.extend([f"{o} (metadata mismatch)" for o in changed_views])

        # compare ddls
        log.info(
            f"COMPARE DDLS (objects other than tables) ON target db [ '{config.database_target}' ] AND meta db [ '{config.database_meta}' ]"
        )
        for object_local in solution_client.all_objects:
            if (
                object_local.object_type == DbObjectType.SCHEMA
            ):  # schemas do not need to be compared
                continue
            if (
                object_local.object_type == DbObjectType.STAGE
            ):  # get_ddl not supported in get_ddl, i.e. DDLs cannot be compared
                continue
            elif (
                object_local.object_type == DbObjectType.TABLE
            ):  # tables cannot be compared using DDLs (order of columns might not be the same)
                continue
            else:
                missing_object, changed_object = _compare_object_ddl(
                    object_local, config, snow_client_account, detailed_output
                )
                missing_objects.extend(missing_object)
                changed_objects.extend([f"{o} (DDL mismatch)" for o in changed_object])

        _print_summary(
            missing_objects,
            changed_objects,
            config.project_folder,
            config.database_target,
        )

    except Exception as err:
        log.error(str(err))
        raise err

    finally:
        database_management.drop_database_and_set_retention_time(
            config.database_meta, snow_client_account, config.temp_db_retention_time
        )


def _compare_tables_and_views_metadata(
    snow_client_meta: SnowClient,
    snow_client_target: SnowClient,
    config: SolutionConfig,
    ignore_column_order=True,
    snow_edition: str = "Enterprise",
):
    """
        Compares tables and views by their metadata
    Returns:
        (table_names_missing, table_names_alter), where
            table_names_missing is a list containing n strings (object name)
            table_names_alter is a list containing n strings (object name)
    """
    metadata_service_metadb = MetadataService(snow_client_meta, snow_edition, config.disabled_object_types)
    metadata_service_metadb.get_all_metadata(
        config.schema_list, config.parallel_threads
    )
    metadata_service_targetdb = MetadataService(snow_client_target, snow_edition, config.disabled_object_types)
    metadata_service_targetdb.get_all_metadata(
        config.schema_list, config.parallel_threads
    )

    def compare(meta_objects_list, target_objects_list):
        object_names_alter = []
        object_names_missing = []
        for meta_object in meta_objects_list:
            missing = True
            if meta_object.schema.upper() == "INFORMATION_SCHEMA":
                continue
            for target_object in target_objects_list:
                target_object.ignore_column_order = ignore_column_order
                if (target_object.schema == meta_object.schema) and (
                    target_object.name == meta_object.name
                ):
                    if target_object != meta_object:
                        log.info(
                            f"METADATA MISMATCH [ '{meta_object.schema}.{meta_object.name}' ] ON databases [ '{config.database_target}' ] and [ '{config.database_meta}' ]"
                        )
                        log.info(
                            InstanceTable.generate_diff_description(
                                target_object, meta_object
                            )
                        )
                        object_names_alter.append(
                            f"{meta_object.schema}.{meta_object.name}"
                        )
                    missing = False
                    break
            if missing:
                log.info(
                    f"OBJECT NOT FOUND [ '{meta_object.schema}.{meta_object.name}' ] ON database [ '{config.database_target}' ]"
                )
                object_names_missing.append(f"{meta_object.schema}.{meta_object.name}")
        return object_names_missing, object_names_alter

    table_names_missing, table_names_alter = compare(
        metadata_service_metadb.tables, metadata_service_targetdb.tables
    )
    view_names_missing, view_names_alter = compare(
        metadata_service_metadb.views, metadata_service_targetdb.views
    )

    return table_names_missing, table_names_alter, view_names_missing, view_names_alter


def _compare_object_ddl(
    object_local, config, snow_client_account, detailed_output=False
):
    """
        Compares object by their DDLs
    Returns:
        (missing_object, changed_object), where
            missing_object is a list containing 1 or 0 strings (object name)
            changed_object is a list containing 1 or 0 strings (object name)
    """
    ddl_target = _get_object_ddl(
        config.database_target,
        object_local.full_name,
        object_local.object_type,
        snow_client_account,
    )
    if not ddl_target:
        log.debug(
            f"DDL NOT FOUND [ '{object_local.full_name}' ] ON database [ '{config.database_target}' ]"
        )
        return [object_local.full_name], []

    ddl_target = _remove_whitespace(string_util.remove_comment(ddl_target))
    ddl_meta = _get_object_ddl(
        config.database_meta,
        object_local.full_name,
        object_local.object_type,
        snow_client_account,
    )
    ddl_meta = _remove_whitespace(string_util.remove_comment(ddl_meta))

    if object_local.object_type in (DbObjectType.PROCEDURE, DbObjectType.FUNCTION):
        ddl_target = _remove_whitespace(
            _remove_comment_from_procedure_function_body(ddl_target)
        )
        ddl_meta = _remove_whitespace(
            _remove_comment_from_procedure_function_body(ddl_meta)
        )
    if object_local.object_type in (DbObjectType.TASK,):
        ddl_target = ddl_target.replace(f"{config.database_target}.", "")
        ddl_meta = ddl_meta.replace(f"{config.database_meta}.", "")

    if ddl_meta.casefold().strip() != ddl_target.casefold().strip():
        if detailed_output:
            log.info(
                f"DDL MISMATCH [ '{object_local.full_name}' ] ON databases [ '{config.database_target}' ] and [ '{config.database_meta}' ]"
            )
            log.info(
                "".join(
                    difflib.unified_diff(
                        ddl_meta.splitlines(keepends=True),
                        ddl_target.splitlines(keepends=True),
                        fromfile="solution",
                        tofile="database",
                    )
                )
            )
        else:
            log.debug(
                f"DDL MISMATCH [ '{object_local.full_name}' ] ON databases [ '{config.database_target}' ] and [ '{config.database_meta}' ]"
            )
        return [], [object_local.full_name]

    return [], []


def _remove_comment_from_procedure_function_body(ddl):
    """
    Remove comments from the body of a function or procedure.

    Comments in the body of a function or procedure might not be automatically removed if they are given in single quotes.
    This function removes the comments and returns the result.
    Input should have comments (outside of body) already removed.
    """
    try:
        body = re.search(
            r"CREATE\s+(?:OR\s+REPLACE\s+)(?:PROCEDURE|FUNCTION).*AS\s*'(?P<body>.*)'\s*;",
            ddl,
            re.IGNORECASE + re.DOTALL,
        ).group("body")
        body_clean = string_util.remove_comment(body.replace("''", "'"))
        return ddl.replace(body, body_clean.replace("'", "''"))
    except:
        log.debug(
            f"Could not parse body. Return DDL with comments in body. DDL:\n{ddl}"
        )
        return ddl


def _print_summary(missing_objects, changed_objects, project_folder, target_database):
    message = f"SUMMARY comparison solution [ '{project_folder}' ] and database [ '{target_database}' ]"
    if len(missing_objects) == 0:
        message = message + "\n" + "All objects in solution are on database."
    else:
        message = (
            message
            + "\n"
            + f"List of objects in solution, but not on database (total {len(missing_objects)}):"
        )
        for mo in missing_objects:
            message = message + "\n" + f"    {mo}"
    if len(changed_objects) == 0:
        message = (
            message
            + "\n"
            + "All existing objects on database are identical to objects in solution."
        )
    else:
        message = (
            message
            + "\n"
            + f"List of differences between database and solution (total {len(changed_objects)}):"
        )
        for co in changed_objects:
            message = message + "\n" + f"    {co}"
    message = (
        message
        + "\n"
        + "This function does not check if there are objects on the database which do not appear in the solution."
    )
    message = (
        message
        + "\n"
        + "Objects might not be found on database due to insufficient rights to access the object."
    )
    log.info(message)
    if len(missing_objects) > 0 or len(changed_objects) > 0:
        log.error(
            f"DETECTED [ '{len(missing_objects) + len(changed_objects)}' ] DIFFERENCES between solution [ '{project_folder}' ] and database [ '{target_database}' ]"
        )
        log.info("Scroll up for details on differences.")


def _get_object_ddl(
    database: str,
    full_object_name: str,
    object_type: str,
    snow_client_account: SnowClient,
):
    try:
        object_type_string = DbObjectType.get_object_type_for_get_ddl(object_type)
        query = f"SELECT GET_DDL('{object_type_string}', '{database}.{full_object_name}') AS DDL"
        ddl_result = snow_client_account.execute_query(query)
    except Exception as err:
        if re.search(
            r"SQL compilation error:(.|\n)*does not exist or not authorized",
            repr(err),
            re.IGNORECASE | re.MULTILINE,
        ):
            return None
        else:
            raise err
    return ddl_result[0]["DDL"]


def _remove_whitespace(t):
    """
    Remove whitespace in each line and completely remove empty lines
    """
    return "\n".join([l.strip() for l in t.splitlines() if l.strip() != ""])



from enum import Enum


class AccountObjectActionType(Enum):
    """
    Enum for valid account operations
    """

    ALTER = "ALTER"
    DROP = "DROP"
    ADD = "ADD"
    REVOKE = "REVOKE"
    GRANT = "GRANT"
    SETTAG = "SETTAG"
    UNSETTAG = "UNSETTAG"


class AccountObjectType(Enum):
    """
    Enum for valid account objects
    """

    STORAGEINTEGRATION = "STORAGEINTEGRATION"
    WAREHOUSE = "WAREHOUSE"
    SHARE = "SHARE"
    DATABASE = "DATABASE"
    EXTERNALVOLUME = "EXTERNALVOLUME"

    @staticmethod
    def get_object_domain_from_object_type(object_type) -> str:
        """
        Get Snowflake object domain by object type.
        """
        mapping_object_type_object_domain = {
            AccountObjectType.STORAGEINTEGRATION: "INTEGRATION",
            AccountObjectType.WAREHOUSE: "WAREHOUSE",
            AccountObjectType.SHARE: "SHARE",
            AccountObjectType.DATABASE: "DATABASE",
            AccountObjectType.EXTERNALVOLUME: "EXTERNAL VOLUME"
        }

        object_domain = mapping_object_type_object_domain[object_type]

        return object_domain
    
    @staticmethod
    def get_supported_externalvolume_storage_location_params() -> list:
        """
        Supported parameters for the storage locations of external volumes.
        Note: encryption_kms_key_id is covered by the encryption_type.
        """
        return  [
        'name',
        'storage_provider',
        'storage_base_url',
        'storage_aws_role_arn',
        'storage_aws_external_id',
        'azure_tenant_id',
        'encryption_type'
        ]


import os

import setuptools

version_id = os.getenv("BUILD_BUILDNUMBER", "0.0.1")

setuptools.setup(
    version=version_id,
    packages=setuptools.find_namespace_packages(),
    include_package_data=True,
    zip_safe=False,
)



import logging
import textwrap
from datetime import datetime
from enum import Enum

from snowflake.connector import SnowflakeConnection


class LogStatus(Enum):
    """
    Indicates which status a deployment step has
    """

    SUCCESS = "SUCCESS"
    FAILED = "FAILED"
    ABORTED = "ABORTED"
    PENDING = "PENDING"
    RUNNING = "RUNNING"
    WARNING = "WARNING"


class LogOperation(Enum):
    """
    Indicates which operational step category a log message falls into.
    """

    DIFF = "DIFF"
    GIT = "GIT"
    SQLCOMMAND = "SQLCOMMAND"
    DEPLOYMENT = "DEPLOYMENT"
    TARGET = "TARGET"
    PREP = "PREPERATION"


class LoggingAdapter(logging.LoggerAdapter):
    """
    Logging adapter for Acedeploy framework.
    Adds the additional args 'operation', 'status', 'db' to the log message.
    """

    def __init__(self, logger, extra=None):
        if extra is None:
            extra = {}
        super().__init__(logger, extra)

    def process(self, msg, kwargs):
        extra_args = ("operation", "status", "db")
        extras = {}
        args = {}
        for kwarg in kwargs:
            if kwarg in extra_args:
                extras[kwarg] = kwargs[kwarg]
            else:
                args[kwarg] = kwargs[kwarg]
        args["extra"] = extras
        return msg, args


class LogFileFormatter(logging.Formatter):
    """
    Logging formatter for saving logfiles.
    Adds the additional args 'operation', 'status', 'db' to the log message.
    Note: default syntax logger.info("my message %s", "some string") is not supported by this
    """

    def format(self, record):
        timestring = datetime.utcfromtimestamp(record.created).isoformat(
            sep=" ", timespec="milliseconds"
        )
        result = f"{timestring} - {record.name} - {record.levelname} - {record.msg}"
        extra_args = ("operation", "status", "db")
        for arg in extra_args:
            if hasattr(record, arg):
                result += f" - {arg}: {getattr(record, arg)}"
        return result.replace("\r", "\\r").replace("\n", "\\n")


class DevOpsFormatter(logging.Formatter):
    """
    Logging formatter for use in DevOps pipelines.
    Adds syntax for errors and warnings.
    Note: default syntax logger.info("my message %s", "some string") is not supported by this
    """

    def __init__(self, add_timestamp=False):
        self.add_timestamp = add_timestamp
        super().__init__()

    def format(self, record):
        message = record.msg
        record.msg = ""
        if self.add_timestamp:
            timestring = datetime.utcfromtimestamp(record.created).isoformat(
                sep=" ", timespec="milliseconds"
            )
            header = f"{timestring} - {record.levelname} - "
        else:
            header = f"{record.levelname} - "
        msg = textwrap.indent(message, " " * len(header)).strip()
        record.msg = message
        result = header + msg
        if record.levelno >= logging.ERROR:
            return f"##vso[task.logissue type=error;]{result}"  # \n##vso[task.complete result=Failed;]DONE'
        elif record.levelno >= logging.WARNING:
            return f"##vso[task.logissue type=warning;]{result}"
        else:
            return result


class GitLabFormatter(logging.Formatter):
    """
    Logging formatter for use in terminals.
    Properly indents multiline strings.
    """

    def __init__(self):
        super().__init__()

    def format(self, record):
        timestring = datetime.utcfromtimestamp(record.created).isoformat(
            sep=" ", timespec="milliseconds"
        )
        message = record.msg
        record.msg = ""
        header = f"{timestring} - {record.levelname} - "
        msg = textwrap.indent(message, " " * len(header)).strip()
        record.msg = message
        return header + msg


class DefaultFormatter(logging.Formatter):
    """
    Logging formatter for use in terminals.
    Properly indents multiline strings.
    """

    def __init__(self):
        super().__init__()

    def format(self, record):
        message = record.msg
        record.msg = ""
        header = f"{record.levelname} - "
        msg = textwrap.indent(message, " " * len(header)).strip()
        record.msg = message
        return header + msg


def get_devops_adapter(logger, debug=False, add_timestamp=False) -> LoggingAdapter:
    """
        Convinience function for use in DevOps pipelines.
    Args:
        logger: logging.Logger
        debug: bool - Set output level to DEBUG, else INFO
        add_timestamp: bool - Add timestamp to each entry
    Returns:
        adapter: logging.LoggerAdapter - Adapter to log against
    Example:
        logger = logging.getLogger('acedeploy')
        log = get_devops_adapter(logger, os.getenv('SYSTEM_DEBUG', False), False)
    """
    logger.setLevel(logging.DEBUG)
    log_level = logging.DEBUG if debug else logging.INFO

    sh = logging.StreamHandler()
    sh.setLevel(log_level)
    sh.setFormatter(DevOpsFormatter(add_timestamp))
    logger.addHandler(sh)

    adapter = LoggingAdapter(logger)
    return adapter


def get_gitlab_adapter(logger, debug=False, add_timestamp=False) -> LoggingAdapter:
    """
        Convinience function for use in Gitlab pipelines.
    Args:
        logger: logging.Logger
        debug: bool - Set output level to DEBUG, else INFO
        add_timestamp: bool - Add timestamp to each entry
    Returns:
        adapter: logging.LoggerAdapter - Adapter to log against
    Example:
        logger = logging.getLogger('acedeploy')
        log = get_devops_adapter(logger, os.getenv('DEBUG_ACEDEPLOY', False), False)
    """
    logger.setLevel(logging.DEBUG)
    log_level = logging.DEBUG if debug else logging.INFO

    sh = logging.StreamHandler()
    sh.setLevel(log_level)
    sh.setFormatter(GitLabFormatter(add_timestamp))
    logger.addHandler(sh)

    adapter = LoggingAdapter(logger)
    return adapter


class SnowflakeTableHandler(logging.Handler):
    """
    Logging handler to log to an existing table in Snowflkae
    """

    def __init__(self, sf_connection: SnowflakeConnection, log_table: str, run_id: str):
        """
        Params:
            sf_connection: SnowflakeConnection - snowflake connection with role and warehouse set
            log_table: str - fully qualified name of logging table
            run_id: str - unique identifier for this script run
        """
        super().__init__()
        self.sf_connection = sf_connection
        self.log_table = log_table
        self.run_id = run_id
        self.log_msg = ""

    def emit(self, record):
        timestring = datetime.utcfromtimestamp(record.created).isoformat(
            sep=" ", timespec="milliseconds"
        )
        self.log_msg = record.msg
        self.log_msg = self.log_msg.strip()
        self.log_msg = self.log_msg.replace("'", "''")
        statement = f"""
            INSERT INTO {self.log_table} (
                RUN_ID,
                TIMESTAMP,
                LEVEL,
                MESSAGE,
                DATABASE,
                OPERATION,
                STATUS
            ) VALUES (
                '{self.run_id}',
                '{timestring}',
                '{record.levelname}',
                '{self.log_msg}',
                '{getattr(record, 'db', '')}',
                '{getattr(record, 'operation', '')}',
                '{getattr(record, 'status', '')}'
            )
        """
        try:
            self.sf_connection.execute_string(statement)
        except Exception as err:
            print(f"Error in SnowflakeTableHandler {str(err)}")



from enum import Enum


class DbActionType(Enum):
    """
    Enum for valid database operations
    """

    ALTER = "ALTER"
    DROP = "DROP"
    ADD = "ADD"
    DROPOVERLOADED = "DROPOVERLOADED"


class DbObjectType(Enum):
    """
    Enum for valid database objects
    """

    SCHEMA = "SCHEMA"
    TABLE = "TABLE"
    EXTERNALTABLE = "EXTERNALTABLE"
    VIEW = "VIEW"
    MATERIALIZEDVIEW = "MATERIALIZEDVIEW"
    FUNCTION = "FUNCTION"
    PROCEDURE = "PROCEDURE"
    STAGE = "STAGE"
    FILEFORMAT = "FILEFORMAT"
    STREAM = "STREAM"
    TASK = "TASK"
    PIPE = "PIPE"
    SEQUENCE = "SEQUENCE"
    MASKINGPOLICY = "MASKINGPOLICY"
    ROWACCESSPOLICY = "ROWACCESSPOLICY"
    DYNAMICTABLE = "DYNAMICTABLE"
    HYBRIDTABLE = "HYBRIDTABLE"
    NETWORKRULE = "NETWORKRULE"
    TAG = "TAG"

    @staticmethod
    def get_sql_object_type(db_object) -> str:
        """
        Get SQL identifier for DbObjectType.
        Example: get_sql_object_type(DbObjectType.MATERIALIZEDVIEW) returns "MATERIALIZED VIEW"
        """
        if db_object == DbObjectType.SCHEMA:
            return "SCHEMA"
        if db_object == DbObjectType.TABLE:
            return "TABLE"
        if db_object == DbObjectType.EXTERNALTABLE:
            return "EXTERNAL TABLE"
        if db_object == DbObjectType.VIEW:
            return "VIEW"
        if db_object == DbObjectType.MATERIALIZEDVIEW:
            return "MATERIALIZED VIEW"
        if db_object == DbObjectType.FUNCTION:
            return "FUNCTION"
        if db_object == DbObjectType.PROCEDURE:
            return "PROCEDURE"
        if db_object == DbObjectType.STAGE:
            return "STAGE"
        if db_object == DbObjectType.FILEFORMAT:
            return "FILE FORMAT"
        if db_object == DbObjectType.STREAM:
            return "STREAM"
        if db_object == DbObjectType.TASK:
            return "TASK"
        if db_object == DbObjectType.PIPE:
            return "PIPE"
        if db_object == DbObjectType.SEQUENCE:
            return "SEQUENCE"
        if db_object == DbObjectType.MASKINGPOLICY:
            return "MASKING POLICY"
        if db_object == DbObjectType.ROWACCESSPOLICY:
            return "ROW ACCESS POLICY"
        if db_object == DbObjectType.DYNAMICTABLE:
            return "DYNAMIC TABLE"
        if db_object == DbObjectType.HYBRIDTABLE:
            return "HYBRID TABLE"
        if db_object == DbObjectType.NETWORKRULE:
            return "NETWORK RULE"
        if db_object == DbObjectType.TAG:
            return "TAG"
        else:
            raise ValueError("Given DbObjectType not recognized.")

    @staticmethod
    def get_object_type_for_show(db_object) -> str:
        """
        Get SQL text identifier for use in SHOW commands. https://docs.snowflake.com/en/sql-reference/sql/show.html
        Example: get_object_type_for_show(DbObjectType.MATERIALIZEDVIEW) returns "MATERIALIZED VIEWS"
        """
        if db_object == DbObjectType.SCHEMA:
            return "SCHEMAS"
        if db_object == DbObjectType.TABLE:
            return "TABLES"
        if db_object == DbObjectType.EXTERNALTABLE:
            return "EXTERNAL TABLES"
        if db_object == DbObjectType.VIEW:
            return "VIEWS"
        if db_object == DbObjectType.MATERIALIZEDVIEW:
            return "MATERIALIZED VIEWS"
        if db_object == DbObjectType.FUNCTION:
            return "FUNCTIONS"
        if db_object == DbObjectType.PROCEDURE:
            return "PROCEDURES"
        if db_object == DbObjectType.STAGE:
            return "STAGES"
        if db_object == DbObjectType.FILEFORMAT:
            return "FILE FORMATS"
        if db_object == DbObjectType.STREAM:
            return "STREAMS"
        if db_object == DbObjectType.TASK:
            return "TASKS"
        if db_object == DbObjectType.PIPE:
            return "PIPES"
        if db_object == DbObjectType.SEQUENCE:
            return "SEQUENCES"
        if db_object == DbObjectType.MASKINGPOLICY:
            return "MASKING POLICIES"
        if db_object == DbObjectType.ROWACCESSPOLICY:
            return "ROW ACCESS POLICIES"
        if db_object == DbObjectType.DYNAMICTABLE:
            return "DYNAMIC TABLES"
        if db_object == DbObjectType.HYBRIDTABLE:
            return "HYBRID TABLES"
        if db_object == DbObjectType.NETWORKRULE:
            return "NETWORK RULES"
        if db_object == DbObjectType.TAG:
            return "TAGS"
        else:
            raise ValueError("Given DbObjectType not recognized.")

    @staticmethod
    def get_object_type_for_get_ddl(db_object) -> str:
        """
        Get SQL text identifier for use in GET_DDL(). https://docs.snowflake.com/en/sql-reference/functions/get_ddl.html
        Example: get_object_type_for_get_ddl(DbObjectType.MATERIALIZEDVIEW) returns "VIEW"
        """
        if db_object == DbObjectType.SCHEMA:
            return "SCHEMA"
        if db_object == DbObjectType.TABLE:
            return "TABLE"
        if db_object == DbObjectType.EXTERNALTABLE:
            return "TABLE"
        if db_object == DbObjectType.VIEW:
            return "VIEW"
        if db_object == DbObjectType.MATERIALIZEDVIEW:
            return "VIEW"
        if db_object == DbObjectType.FUNCTION:
            return "FUNCTION"
        if db_object == DbObjectType.PROCEDURE:
            return "PROCEDURE"
        if db_object == DbObjectType.STAGE:
            raise ValueError("Object of type STAGE cannot be used in GET_DDL()")
        if db_object == DbObjectType.FILEFORMAT:
            return "FILE_FORMAT"
        if db_object == DbObjectType.STREAM:
            return "STREAM"
        if db_object == DbObjectType.TASK:
            return "TASK"
        if db_object == DbObjectType.PIPE:
            return "PIPE"
        if db_object == DbObjectType.SEQUENCE:
            return "SEQUENCE"
        if db_object == DbObjectType.MASKINGPOLICY:
            return "POLICY"
        if db_object == DbObjectType.ROWACCESSPOLICY:
            return "POLICY"
        if db_object == DbObjectType.DYNAMICTABLE:
            return "TABLE"
        if db_object == DbObjectType.HYBRIDTABLE:
            return "TABLE"
        if db_object == DbObjectType.NETWORKRULE:
            raise ValueError("Object of type NETWORK_RULE cannot be used in GET_DDL()")
        if db_object == DbObjectType.TAG:
            return "TAG"
        else:
            raise ValueError("Given DbObjectType not recognized.")
        
    @staticmethod
    def get_object_domain_from_object_type_for_tag_references(object_type) -> str:
        """
        Get Snowflake object domain by object type.
        """
        mapping_object_type_object_domain = {
            DbObjectType.TABLE: "TABLE",
            DbObjectType.VIEW: "TABLE",
            DbObjectType.MATERIALIZEDVIEW: "TABLE",
            DbObjectType.EXTERNALTABLE: "TABLE",
            DbObjectType.DYNAMICTABLE: "TABLE",
            DbObjectType.HYBRIDTABLE: "TABLE",  
        }

        object_domain = mapping_object_type_object_domain[object_type]

        return object_domain
    
    @staticmethod
    def get_object_domain_from_object_type_for_alter_tags(object_type) -> str:
        """
        Get Snowflake object domain by object type.
        """
        mapping_object_type_object_domain = {
            DbObjectType.TABLE: "TABLE",
            DbObjectType.VIEW: "VIEW",
            DbObjectType.MATERIALIZEDVIEW: "VIEW",
            DbObjectType.EXTERNALTABLE: "EXTERNAL TABLE",
            DbObjectType.DYNAMICTABLE: "DYNAMIC TABLE",
            DbObjectType.HYBRIDTABLE: "HYBRID TABLE"
        }

        object_domain = mapping_object_type_object_domain[object_type]

        return object_domain
    
    @staticmethod
    def get_object_type_for_policy_references(ref_entity_domain: str) -> 'DbObjectType':
        """
        Get Snowflake object type by string.
        """
        mapping = {
            "TABLE": DbObjectType.TABLE,  # TODO: validate if dynamic/hybrid tables are included here
            "VIEW": DbObjectType.VIEW,
            "TAG": DbObjectType.TAG,
        }
        return mapping[ref_entity_domain]


class DbFunctionType(Enum):
    """
    Enum for valid function types
    """

    SQL = "SQL"
    JAVASCRIPT = "JAVASCRIPT"


class PolicyType(Enum):
    """
    Valid policy types
    """

    ROWACCESS = "ROWACCESS"
    MASKING = "MASKING"



import logging
from typing import Dict, List, Union

import snowflake.connector
from acedeploy.core.model_configuration import SolutionConfig
from aceutils.logger import LoggingAdapter, LogOperation, LogStatus
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.backends import default_backend


logger = logging.getLogger(__name__)
log = LoggingAdapter(logger)


class SnowClientConfig(object):
    def __init__(
        self,
        account: str,
        user: str,
        password: str = None,
        private_key: str = None,
        private_key_pass: str = None,
        warehouse: str = None,
        role: str = None,
        database: str = None,
        sql_variables: dict = None,
    ):
        self.user = user
        self.password = password
        if private_key is not None:
            self.private_key_bytes = self._format_private_key(private_key, private_key_pass)
        else:
            self.private_key_bytes = None
        self.private_key_pass = private_key_pass
        self.account = account
        self.warehouse = warehouse
        self.role = role
        self.database = database
        self.sql_variables = {} if sql_variables is None else sql_variables

    @staticmethod
    def _format_private_key(private_key: str, private_key_pass: str) -> bytes:
        if not private_key:
            raise ValueError("Private key content is not provided.")
        # Remove all unnecessary whitespace to ensure proper formatting
        private_key = "".join(private_key.split())
        # Check if the key starts and ends correctly
        if not (
            private_key.startswith("-----BEGINENCRYPTEDPRIVATEKEY-----")
            and private_key.endswith("-----ENDENCRYPTEDPRIVATEKEY-----")
        ):
            raise ValueError(
                "Private key must start with '-----BEGIN ENCRYPTED PRIVATE KEY-----' and end with '-----END ENCRYPTED PRIVATE KEY-----'."
            )
        # Extract the Base64 content between header and footer
        base64_content = private_key.replace(
            "-----BEGINENCRYPTEDPRIVATEKEY-----", ""
        ).replace("-----ENDENCRYPTEDPRIVATEKEY-----", "")
        # Insert newlines every 64 characters for proper PEM formatting
        base64_lines = [
            base64_content[i : i + 64] for i in range(0, len(base64_content), 64)
        ]
        # Add headers and footers with correct line breaks
        formatted_pk = (
            "-----BEGIN ENCRYPTED PRIVATE KEY-----\n"
            + "\n".join(base64_lines)
            + "\n-----END ENCRYPTED PRIVATE KEY-----"
        )
        formatted_pk = serialization.load_pem_private_key(
            formatted_pk.encode("utf-8"),
            password=private_key_pass.encode("utf-8"),
            backend=default_backend(),
        )
        private_key_bytes = formatted_pk.private_bytes(
            encoding=serialization.Encoding.DER,
            format=serialization.PrivateFormat.PKCS8,
            encryption_algorithm=serialization.NoEncryption(),
        )
        return private_key_bytes

    def get_connect_info(self) -> Dict:
        connect_info = {
            "account": self.account,
            "user": self.user,
            "password": self.password,
            "private_key": self.private_key_bytes,
            "private_key_pass": self.private_key_pass,
        }

        for opt_prop in ["warehouse", "role", "database"]:
            if getattr(self, opt_prop) is not None:
                connect_info[opt_prop] = getattr(self, opt_prop)

        return connect_info

    @staticmethod
    def get_from_solution_config(
        config: SolutionConfig, database: str = None
    ) -> "SnowClientConfig":
        """
            Convienience wrapper to return a snowflake client config object based of the solution config
        Args:
            config: SolutionConfig
            database: str - The name of the database to use in the snowflake connection (optional)
        Returns:
            SnowClientConfig - snowflake client config
        """
        return SnowClientConfig(
            config.snow_account,
            config.snow_login,
            config.snow_password,
            config.snow_private_key,
            config.snow_private_key_pass,
            config.snow_warehouse,
            config.snow_role,
            database,
            config.sql_variables,
        )


class SnowClient(object):
    """
    Objects that collects all operations and information about the snowflake databases.
    """

    def __init__(self, config: SnowClientConfig):
        """
            Creates a new SnowClient
        Args:
            config: SolutionConfig - stored configuration values
        """
        self.connection = None
        self._config = config
        self.database = self._config.database

    def __del__(self):
        """
        Ensures the connection snowflake is closed at the end
        """
        if self.connection is not None:
            self.connection.close()

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        if self.connection is not None:
            self.connection.close()

    def _get_connection(self):
        """
        Factory method to get snowflake connection
        Initializes new snowflake connection if not already connected
        """
        if self.connection is None:
            conn_info = self._config.get_connect_info()
            self.connection = snowflake.connector.connect(**conn_info)
            if self._config.database is not None:
                self.execute_statement(f"USE DATABASE {self.database};")

            self.set_sql_variables(self._config.sql_variables)

        return self.connection

    @staticmethod
    def _get_error_message(excepction: Exception, statement: str) -> None:
        """
        Compose error message if the execution of a statement or query fails.
        """
        if hasattr(excepction, "raw_msg"):
            message = excepction.raw_msg.replace("\n", " ")
        else:
            message = str(
                excepction
            )  # this makes sure that all kinds of errors can have a message, even if they do not have raw_msg attribute
        if hasattr(excepction, "sfqid"):
            message = message + f"\nQuery ID: {excepction.sfqid}"
        return f"SNOWFLAKE ERROR: {message}\nFailed statement:\n{statement}"

    def execute_statement(self, statement: Union[str, List[str]]) -> None:
        """
            Executes simple statement against snowflake
            Schema and Database settings must be set beforehand
        Args:
            statement Union[str, List[str]] - a sql statement or a list of sql statements to execute
        """
        connection = self._get_connection()
        statement_list: List[str] = (
            statement if isinstance(statement, list) else [statement]
        )

        try:
            for single_statement in statement_list:
                stripped_statement = (
                    single_statement.strip()
                )  # remove whitespace from statement, as execute_string() might produce warnings if empty new lines are found after a semicolon
                log.debug(
                    f"START execution [ '{stripped_statement}' ]",
                    operation=LogOperation.SQLCOMMAND,
                    status=LogStatus.PENDING,
                    db=self.database,
                )
                _ = connection.execute_string(stripped_statement)
                log.debug(
                    f"FINISH execution [ '{stripped_statement}' ]",
                    operation=LogOperation.SQLCOMMAND,
                    status=LogStatus.SUCCESS,
                    db=self.database,
                )

        except Exception as err:
            raise Exception(self._get_error_message(err, single_statement)) from err

    def execute_query(
        self, query: Union[str, List[str]], use_dict_cursor: bool = True
    ) -> Union[List[Dict], List[List[Dict]]]:
        """
            Executes sql statements and against snowflake and returns the result as dictionary or list of dictionaries
        Args:
            query Union[str, List[str]] - a sql query or a list of sql queries to execute
            use_dict_cursor bool (default true) - use snowflake DictCursor for results instead of returning a list
        """
        connection = self._get_connection()
        if use_dict_cursor is True:
            cursor = connection.cursor(snowflake.connector.DictCursor)
        else:
            cursor = connection.cursor()

        query_list: List[str] = query if isinstance(query, list) else [query]
        result: List[Dict] = []

        try:
            for single_query in query_list:
                log.debug(
                    f"START execution [ '{single_query}' ]",
                    operation=LogOperation.SQLCOMMAND,
                    status=LogStatus.PENDING,
                    db=self.database,
                )
                result.append(cursor.execute(single_query).fetchall())
                log.debug(
                    f"FINISH execution [ '{single_query}' ]",
                    operation=LogOperation.SQLCOMMAND,
                    status=LogStatus.SUCCESS,
                    db=self.database,
                )
        except Exception as err:
            raise Exception(self._get_error_message(err, single_query)) from err

        return result[0] if not isinstance(query, list) else result

    def set_sql_variables(self, variables_dict: dict) -> None:
        """
            Set given SQL variables for the session

            Supported variable types: int, float, bool, str
        Args
            variables_dict: dict - dictionary containing the variables to be set, example: {"varname1": "varvalue1", "varname2": "varvalue2"}
        """
        log.debug("SET sql variables")
        for key, value in variables_dict.items():
            if isinstance(value, int):
                value_string = str(value)
            elif isinstance(value, float):
                value_string = str(value)
            elif isinstance(value, bool):
                if value:
                    value_string = "TRUE"
                else:
                    value_string = "FALSE"
            else:
                if "'" in value:
                    raise ValueError(
                        f"Variable [ '{key}' ] contains illegal character \"'\"."
                    )
                value_string = f"'{value}'"
            statement = f"SET {key} = {value_string}"
            self.execute_statement(statement)



import os
from typing import Dict, List, Union

import aceutils.file_util as file_util
from acedeploy.services.secret_service import EnvironmentSecretService


class AccountObjectConfig(object):
    """
    Holds solution config.
    """

    def __init__(self):
        """ """
        self.module_root_folder = os.path.abspath(
            os.path.join(os.path.dirname(__file__), "..")
        )

        self.config_path = os.environ.get("ACEACCOUNT_CONFIG_PATH")

        self.config_dict = file_util.load_json(self.config_path)

        key_service = EnvironmentSecretService()
        self.key_service = key_service

        self.parse_config()

    def parse_config(self):
        """
        Parses config file.
        """
        self.source_files = self._get_env_var(
            self._get_nested_dict_value(self.config_dict, ["sourceFiles"])
        )

        self.snow_edition = self._get_env_var(
            self._get_nested_dict_value(
                self.config_dict, ["connection", "snowflakeEdition"], "Enterprise"
            )
        )
        self.snow_account = self._get_env_var(
            self._get_nested_dict_value(self.config_dict, ["connection", "account"])
        )
        self.snow_login = self._get_env_var(
            self._get_nested_dict_value(self.config_dict, ["connection", "login"])
        )
        self.snow_password = self._get_env_var(
            self._get_nested_dict_value(self.config_dict, ["connection", "password"])
        )
        self.snow_role = self._get_env_var(
            self._get_nested_dict_value(self.config_dict, ["connection", "role"])
        )
        self.snow_warehouse = self._get_env_var(
            self._get_nested_dict_value(self.config_dict, ["connection", "warehouse"])
        )
        self.object_options = self._get_env_var(
            self._get_nested_dict_value(self.config_dict, ["objectOptions"])
        )

        self.drop_enabled = {object_type: object_option["dropEnabled"] for object_type, object_option in self.object_options.items()}

        self.enabled_object_types = [object_type for object_type, object_option in self.object_options.items() if object_option["enabled"]]

        self.revoke_enabled = self._get_env_var(
            self._get_nested_dict_value(self.config_dict, ["revokeEnabled"])
        ) 
        self.unsettag_enabled = self._get_env_var(
            self._get_nested_dict_value(self.config_dict, ["unsettagEnabled"])
        ) 
        self.max_number_of_threads = self._get_env_var(
            self._get_nested_dict_value(self.config_dict, ["maxNumberOfThreads"])
        )        
        self.max_number_of_threads = 1 if self.max_number_of_threads is None else self.max_number_of_threads

    @staticmethod
    def _get_nested_dict_value(
        nested_dict: Dict, keys: List[str], default: str = None
    ) -> str:
        data = nested_dict
        for k in keys:
            if k in data:
                data = data[k]
            else:
                return default
        return data

    def _get_env_var(self, val: str) -> Union[str, bool]:
        """
            If the given value is enclosed with '@@', load value from environment variable of that name.
            Else, return the value.
            If the environment variable value is the string representation of a boolean, return that boolean.
        Args:
            val: str - environment variable name enclosed in '@@', or value
        Raises:
            ValueError - if secret was not found in configured Key Vault
        Returns:
            if value is not marked as name of environment variable: value
            if value is marked as name of environment variable:
                if environment variable is found - will return the environment variable value
                if environment variable is not found - ValueError
        """
        char_delimiter = "@@"

        if (
            isinstance(val, str)
            and val.startswith(char_delimiter)
            and val.endswith(char_delimiter)
        ):
            env_var_name = val[len(char_delimiter) : -len(char_delimiter)]
            env_var_value = self.key_service.get_secret(env_var_name)
            if env_var_value in ["True", "TRUE", "true"]:
                return True
            elif env_var_value in ["False", "FALSE", "false"]:
                return False
            else:
                return env_var_value
        else:
            return val



from aceaccount.core.model_account_object import AccountObject
from aceaccount.core.model_account_object_sql_entities import AccountObjectType

class AccountObjectStatement(AccountObject):
    """
    Models atomic sql statement execution
    With target object name and sql statement to execute
    """

    def __init__(self, name: str, statement: str, object_type: AccountObjectType):
        super().__init__(name=name, object_type=object_type)
        self.statement = statement

    def __eq__(self, other):
        return (
            self.name == other.name
            and self.statement == other.statement
            and self.object_type == other.object_type
        )

    def __str__(self):
        return f"AccountObjectStatement: {self.id}"

    def __repr__(self):
        return f"AccountObjectStatement: {self.id}"